# -*- coding: utf-8 -*-
"""
Objects and functions dealing with seismic phases.
"""
from itertools import count
import math
import re

import numpy as np
from scipy.optimize import brentq

from obspy.core.util.obspy_types import Enum

from .helper_classes import (Arrival, SlownessModelError, TauModelError,
                             TimeDist)

from .c_wrappers import clibtau
from . import _DEFAULT_VALUES


_ACTIONS = Enum([
    # Used by add_to_branch when the path turns within a segment. We assume
    # that no ray will turn downward so turning implies turning from downward
    # to upward, ie U.
    "turn",
    # Used by add_to_branch when the path reflects off the top of the end of
    # a segment, ie ^.
    "reflect_underside",
    # Used by add_to_branch when the path reflects off the bottom of the end
    # of a segment, ie v.
    "reflect_topside",
    # Used by add_to_branch when the path transmits up through the end of a
    # segment.
    "transup",
    # Used by add_to_branch when the path transmits down through the end of a
    # segment.
    "transdown",
    # Used by add_to_branch when the path diffracts
    "diffract"
])


class SeismicPhase(object):
    """
    Stores and transforms seismic phase names to and from their
    corresponding sequence of branches. Will maybe contain "expert" mode
    wherein paths may start in the core. Principal use is to calculate leg
    contributions for scattered phases. Nomenclature: "K" - downgoing wave
    from source in core; "k" - upgoing wave from source in core.
    """
    def __init__(self, name, tau_model, receiver_depth=0.0):
        # The phase name, e.g. PKiKP.
        self.name = name
        # The receiver depth within the TauModel that was used to generate this
        # phase. Normally this is 0.0 for a surface station, but can be
        # different for borehole or scattering calculations.
        self.receiver_depth = receiver_depth
        # TauModel to generate phase for.
        self.tau_model = tau_model

        # The source depth within the TauModel that was used to generate
        # this phase.
        self.source_depth = self.tau_model.source_depth

        # List containing strings for each leg.
        self.legs = leg_puller(name)

        # Name with depths corrected to be actual discontinuities in the model.
        self.purist_name = self.create_purist_name(tau_model)

        # Settings for this instance. Should eventually be configurable.
        self._settings = {
            # The maximum degrees that a Pn or Sn can refract along the moho.
            # Note this is not the total distance, only the segment along the
            # moho.
            "max_refraction_in_radians": np.radians(20.0),
            # The maximum degrees that a Pdiff or Sdiff can diffract along the
            # CMB. Note this is not the total distance, only the segment along
            # the CMB.
            "max_diffraction_in_radians": np.radians(60.0),
            # The maximum number of refinements to make to an Arrival.
            "max_recursion": 50
        }

        # Enables phases originating in core.
        self.expert = False
        # Minimum/maximum ray parameters that exist for this phase.
        self.min_ray_param = None
        self.max_ray_param = None
        # Index within TauModel.ray_param that corresponds to max_ray_param.
        # Note that max_ray_param_index < min_ray_param_index as ray parameter
        # decreases with increasing index.
        self.max_ray_param_index = -1
        # Index within TauModel.ray_param that corresponds to min_ray_param.
        # Note that max_ray_param_index < min_ray_param_index as ray parameter
        # decreases with increasing index.
        self.min_ray_param_index = -1
        # Temporary branch numbers determining where to start adding to the
        # branch sequence.
        self.current_branch = None
        # Array of distances corresponding to the ray parameters stored in
        # ray_param.
        self.dist = None
        # Array of times corresponding to the ray parameters stored in
        # ray_param.
        self.time = None
        # Array of possible ray parameters for this phase.
        self.ray_param = None
        # The minimum distance that this phase can be theoretically observed.
        self.min_distance = 0.0
        # The maximum distance that this phase can be theoretically observed.
        self.max_distance = 1e300
        # List (could make array!) of branch numbers for the given phase.
        # Note that this depends upon both the planet model and the source
        # depth.
        self.branch_seq = []
        # List of branch_seq positions where a head or diffracted
        # segment occurs.
        self.head_or_diffract_seq = []
        # Description of segments of the phase.
        self.segment_list = []
        # True if the current leg of the phase is down going. This allows a
        # check to make sure the path is correct.
        # Used in addToBranch() and parseName().
        self.down_going = []
        # ArrayList of wave types corresponding to each leg of the phase.
        self.wave_type = []

        self.parse_name(tau_model)
        self.sum_branches(tau_model)

    def create_purist_name(self, tau_model):
        current_leg = self.legs[0]
        # Deal with surface wave veocities first, since they are a special
        # case.
        if len(self.legs) == 2 and current_leg.endswith("kmps"):
            purist_name = self.name
            return purist_name
        purist_name = ""
        # Only loop to penultimate element as last leg is always "END".
        for current_leg in self.legs[:-1]:
            # Find out if the next leg represents a phase conversion or
            # reflection depth.
            if current_leg[0] in "v^":
                discon_branch = closest_branch_to_depth(tau_model,
                                                        current_leg[1:])
                leg_depth = tau_model.tau_branches[0, discon_branch].top_depth
                purist_name += current_leg[0]
                purist_name += str(int(round(leg_depth)))
            else:
                try:
                    float(current_leg)
                except ValueError:
                    # If current_leg is just a string:
                    purist_name += current_leg
                else:
                    # If it is indeed a number:
                    discon_branch = closest_branch_to_depth(tau_model,
                                                            current_leg)
                    leg_depth = \
                        tau_model.tau_branches[0, discon_branch].top_depth
                    purist_name += str(leg_depth)
        return purist_name

    def parse_name(self, tau_model):
        """
        Construct a branch sequence from the given phase name and tau model.
        """
        current_leg = self.legs[0]
        next_leg = current_leg

        # Deal with surface wave velocities first, since they are a special
        # case.
        if len(self.legs) == 2 and current_leg.endswith("kmps"):
            return

        # Make a check for J legs if the model doesn't allow J:
        if "J" in self.name and not tau_model.s_mod.allow_inner_core_s:
            raise TauModelError("J phases are not created for this model: {}"
                                .format(self.name))

        # Set currWave to be the wave type for this leg, P or S
        if current_leg in ("p", "k", "I") or current_leg[0] in ("P", "K"):
            is_p_wave = True
            is_p_wave_previous = is_p_wave
        elif current_leg in ("s", "J") or current_leg[0] == "S":
            is_p_wave = False
            is_p_wave_previous = is_p_wave
        else:
            raise TauModelError('Unknown starting phase: ' + current_leg)

        # First, decide whether the ray is upgoing or downgoing from the
        # source. If it is up going then the first branch number would be
        # model.source_branch-1 and downgoing would be model.source_branch.
        upgoing_rec_branch = tau_model.find_branch(self.receiver_depth)
        downgoing_rec_branch = upgoing_rec_branch - 1  # One branch shallower.
        if current_leg[0] in "sS":
            # Exclude S sources in fluids.
            sdep = tau_model.source_depth
            if tau_model.cmb_depth < sdep < tau_model.iocb_depth:
                self.max_ray_param, self.min_ray_param = -1, -1
                return

        # Set self.max_ray_param to be a horizontal ray leaving the source and
        # self.min_ray_param to be a vertical (p=0) ray.
        if current_leg[0] in "PS" or (self.expert and current_leg[0] in "KIJ"):
            # Downgoing from source.
            self.current_branch = tau_model.source_branch
            # Treat initial downgoing as if it were an underside reflection.
            end_action = _ACTIONS["reflect_underside"]
            try:
                s_layer_num = tau_model.s_mod.layer_number_below(
                    tau_model.source_depth, is_p_wave_previous)
                layer = tau_model.s_mod.get_slowness_layer(
                    s_layer_num, is_p_wave_previous)
                self.max_ray_param = layer['top_p']
            except SlownessModelError as e:
                msg = ('Please contact the developers. This error should not '
                       'occur.')
                raise RuntimeError(msg) from e
            self.max_ray_param = tau_model.get_tau_branch(
                tau_model.source_branch, is_p_wave).max_ray_param
        elif current_leg in ("p", "s") or (
                self.expert and current_leg[0] == "k"):
            # Upgoing from source: treat initial downgoing as if it were a
            # topside reflection.
            end_action = _ACTIONS["reflect_topside"]
            try:
                s_layer_num = tau_model.s_mod.layer_number_above(
                    tau_model.source_depth, is_p_wave_previous)
                layer = tau_model.s_mod.get_slowness_layer(
                    s_layer_num, is_p_wave_previous)
                self.max_ray_param = layer['bot_p']
                #  check if source is in high slowness zone
                in_high_slowness, high_slowness_range = \
                    tau_model.s_mod.depth_in_high_slowness(
                        tau_model.source_depth,
                        self.max_ray_param,
                        is_p_wave_previous,
                        return_depth_range=True)
                if in_high_slowness:
                    # need to reduce max_ray_param until it can propagate out
                    # of high slowness zone
                    self.max_ray_param = min(
                        self.max_ray_param,
                        high_slowness_range.ray_param)

            except SlownessModelError as e:
                msg = ('Please contact the developers. This error should not '
                       'occur.')
                raise RuntimeError(msg) from e
            if tau_model.source_branch != 0:
                self.current_branch = tau_model.source_branch - 1
            else:
                # p and s for zero source depth are only at zero distance
                # and then can be called P or S.
                self.max_ray_param = -1
                self.min_ray_param = -1
                return
        else:
            raise TauModelError(
                'First phase not recognised {}: Must be one of P, Pg, Pn, '
                'Pdiff, p, Ped or the S equivalents.'.format(current_leg))
        if self.receiver_depth != 0:
            if self.legs[-2] in ('Ped', 'Sed'):
                # Downgoing at receiver
                self.max_ray_param = min(
                    tau_model.get_tau_branch(
                        downgoing_rec_branch, is_p_wave).min_turn_ray_param,
                    self.max_ray_param)
            else:
                # upgoing at receiver
                self.max_ray_param = min(
                    tau_model.get_tau_branch(upgoing_rec_branch,
                                             is_p_wave).max_ray_param,
                    self.max_ray_param)

        self.min_ray_param = 0

        is_leg_depth, is_next_leg_depth = False, False

        # Now loop over all the phase legs and construct the proper branch
        # sequence.
        current_leg = "START"  # So the prev_leg isn't wrong on the first pass.
        for leg_num in range(len(self.legs) - 1):
            prev_leg = current_leg
            current_leg = next_leg
            next_leg = self.legs[leg_num + 1]
            is_leg_depth = is_next_leg_depth

            # Find out if the next leg represents a phase conversion depth.
            try:
                next_leg_depth = float(next_leg)
                is_next_leg_depth = True
            except ValueError:
                next_leg_depth = -1
                is_next_leg_depth = False

            # Set currWave to be the wave type for this leg, "P" or "S".
            is_p_wave_previous = is_p_wave
            if current_leg in ("p", "k", "I") or current_leg[0] == "P":
                is_p_wave = True
            elif current_leg in ("s", "J") or current_leg[0] == "S":
                is_p_wave = False
            elif current_leg[0] == "K":
                # Here we want to use whatever is_p_wave was on the last leg
                # so do nothing. This makes sure we use the correct
                # max_ray_param from the correct TauBranch within the outer
                # core. In other words K has a high slowness zone if it
                # entered the outer core as a mantle P wave, but doesn't if
                # it entered as a mantle S wave. It shouldn't matter for
                # inner core to outer core type legs.
                pass

            # Check to see if there has been a phase conversion.
            if len(self.branch_seq) > 0 and is_p_wave_previous != is_p_wave:
                self.phase_conversion(tau_model, self.branch_seq[-1],
                                      end_action, is_p_wave_previous)

            if current_leg in ('Ped', 'Sed'):
                if next_leg == "END":
                    if self.receiver_depth > 0:
                        end_action = _ACTIONS["reflect_underside"]
                        self.add_to_branch(tau_model, self.current_branch,
                                           downgoing_rec_branch, is_p_wave,
                                           end_action)
                    else:
                        # This should be impossible except for 0 dist 0 source
                        # depth which can be called p or P.
                        self.max_ray_param = -1
                        self.min_ray_param = -1
                        return
                elif next_leg in ('Pdiff', 'Sdiff'):
                    end_action = _ACTIONS["diffract"]
                    self.add_to_branch(tau_model, self.current_branch,
                                       tau_model.cmb_branch - 1, is_p_wave,
                                       end_action)
                    if current_leg[0] != next_leg[0]:
                        # like Sed Pdiff conversion
                        is_p_wave = not is_p_wave
                elif next_leg[0] == "K":
                    end_action = _ACTIONS["transdown"]
                    self.add_to_branch(tau_model, self.current_branch,
                                       tau_model.cmb_branch - 1, is_p_wave,
                                       end_action)
                elif next_leg == "m":
                    end_action = _ACTIONS["transdown"]
                    self.add_to_branch(tau_model, self.current_branch,
                                       tau_model.moho_branch - 1, is_p_wave,
                                       end_action)
                elif next_leg.startswith("v"):
                    discon_branch = closest_branch_to_depth(
                        tau_model, next_leg[1:])
                    if self.current_branch <= discon_branch - 1:
                        end_action = _ACTIONS["reflect_topside"]
                        self.add_to_branch(
                            tau_model, self.current_branch, discon_branch - 1,
                            is_p_wave, end_action)
                    else:
                        raise TauModelError(
                            "Phase not recognized: {} followed by {}".format(
                                current_leg, next_leg))
                elif next_leg in ('I', 'J'):
                    if tau_model.cmb_depth == tau_model.iocb_depth:
                        # degenerate case of no fluid outer core,
                        # so allow phases like PIP or SJS
                        end_action = _ACTIONS["transdown"]
                        self.add_to_branch(
                            tau_model, self.current_branch,
                            tau_model.cmb_branch - 1,
                            is_p_wave, end_action)
                    else:
                        self.max_ray_param = -1
                        return
                else:
                    raise TauModelError(
                        "Phase not recognized: {} followed by {}".format(
                            current_leg, next_leg))

            # Deal with p and s case.
            elif current_leg in ("p", "s", "k"):
                if next_leg[0] == "v":
                    raise TauModelError(
                        "p and s must always be upgoing and cannot come "
                        "immediately before a top-sided reflection.")
                elif next_leg.startswith("^"):
                    discon_branch = closest_branch_to_depth(
                        tau_model, next_leg[1:])
                    if self.current_branch >= discon_branch:
                        end_action = _ACTIONS["reflect_underside"]
                        self.add_to_branch(
                            tau_model, self.current_branch, discon_branch,
                            is_p_wave, end_action)
                    else:
                        raise TauModelError(
                            "Phase not recognised: {} followed by {} when "
                            "current_branch > discon_branch".format(
                                current_leg, next_leg))
                elif next_leg == "m" and \
                        self.current_branch >= tau_model.moho_branch:
                    end_action = _ACTIONS["transup"]
                    self.add_to_branch(
                        tau_model, self.current_branch, tau_model.moho_branch,
                        is_p_wave, end_action)
                elif next_leg[0] in ("P", "S", "K") or next_leg == "END":
                    if next_leg == 'END':
                        discon_branch = upgoing_rec_branch
                    elif next_leg[0] == 'K':
                        discon_branch = tau_model.cmb_branch
                    else:
                        discon_branch = 0
                    if current_leg == 'k' and next_leg[0] != 'K':
                        end_action = _ACTIONS["transup"]
                    else:
                        end_action = _ACTIONS["reflect_underside"]
                    self.add_to_branch(
                        tau_model, self.current_branch, discon_branch,
                        is_p_wave, end_action)
                elif is_next_leg_depth:
                    discon_branch = closest_branch_to_depth(tau_model,
                                                            next_leg)
                    end_action = _ACTIONS["transup"]
                    self.add_to_branch(
                        tau_model, self.current_branch, discon_branch,
                        is_p_wave, end_action)
                else:
                    raise TauModelError(
                        "Phase not recognized: {} followed by {}".format(
                            current_leg, next_leg))

            # Now deal with P and S case.
            elif current_leg in ("P", "S"):
                if next_leg in ("P", "S", "Pn", "Sn", "Pdiff", "Sdiff", "END"):
                    if end_action == _ACTIONS["transdown"] or \
                            end_action == _ACTIONS["reflect_underside"]:
                        # Was downgoing, so must first turn in mantle.
                        end_action = _ACTIONS["turn"]
                        self.add_to_branch(tau_model, self.current_branch,
                                           tau_model.cmb_branch - 1, is_p_wave,
                                           end_action)
                    if next_leg == 'END':
                        end_action = _ACTIONS["reflect_underside"]
                        self.add_to_branch(tau_model, self.current_branch,
                                           upgoing_rec_branch, is_p_wave,
                                           end_action)
                    else:
                        end_action = _ACTIONS["reflect_underside"]
                        self.add_to_branch(
                            tau_model, self.current_branch, 0, is_p_wave,
                            end_action)
                elif next_leg[0] == "v":
                    discon_branch = closest_branch_to_depth(tau_model,
                                                            next_leg[1:])
                    if self.current_branch <= discon_branch - 1:
                        end_action = _ACTIONS["reflect_topside"]
                        self.add_to_branch(tau_model, self.current_branch,
                                           discon_branch - 1, is_p_wave,
                                           end_action)
                    else:
                        raise TauModelError(
                            "Phase not recognised: {} followed by {} when "
                            "current_branch > discon_branch".format(
                                current_leg, next_leg))
                elif next_leg[0] == "^":
                    discon_branch = closest_branch_to_depth(tau_model,
                                                            next_leg[1:])
                    if prev_leg[0] == "K":
                        end_action = _ACTIONS["reflect_underside"]
                        self.add_to_branch(tau_model, self.current_branch,
                                           discon_branch, is_p_wave,
                                           end_action)
                    elif prev_leg[0] == "^" or prev_leg in ("P", "S", "p", "s",
                                                            "START"):
                        end_action = _ACTIONS["turn"]
                        self.add_to_branch(tau_model, self.current_branch,
                                           tau_model.cmb_branch - 1, is_p_wave,
                                           end_action)
                        end_action = _ACTIONS["reflect_underside"]
                        self.add_to_branch(
                            tau_model, self.current_branch, discon_branch,
                            is_p_wave, end_action)
                    elif ((prev_leg[0] == "v" and
                            discon_branch < closest_branch_to_depth(
                                tau_model, prev_leg[1:]) or
                           (prev_leg == "m" and
                               discon_branch < tau_model.moho_branch) or
                           (prev_leg == "c" and
                               discon_branch < tau_model.cmb_branch))):
                        end_action = _ACTIONS["reflect_underside"]
                        self.add_to_branch(
                            tau_model, self.current_branch, discon_branch,
                            is_p_wave, end_action)
                    else:
                        raise TauModelError(
                            "Phase not recognised: {} followed by {} when "
                            "current_branch > discon_branch".format(
                                current_leg, next_leg))
                elif next_leg == "c":
                    end_action = _ACTIONS["reflect_topside"]
                    self.add_to_branch(
                        tau_model, self.current_branch,
                        tau_model.cmb_branch - 1, is_p_wave, end_action)
                elif next_leg[0] == "K":
                    end_action = _ACTIONS["transdown"]
                    self.add_to_branch(
                        tau_model, self.current_branch,
                        tau_model.cmb_branch - 1, is_p_wave, end_action)
                elif next_leg == "m" or (is_next_leg_depth and
                                         next_leg_depth < tau_model.cmb_depth):
                    # Treat the Moho in the same way as 410 type
                    # discontinuities.
                    discon_branch = closest_branch_to_depth(tau_model,
                                                            next_leg)
                    if end_action == _ACTIONS["turn"] \
                            or end_action == _ACTIONS["reflect_topside"] \
                            or end_action == _ACTIONS["transup"]:
                        # Upgoing section
                        if discon_branch > self.current_branch:
                            # Check the discontinuity below the current
                            # branch when the ray should be upgoing
                            raise TauModelError(
                                "Phase not recognised: {} followed by {} when "
                                "current_branch > discon_branch".format(
                                    current_leg, next_leg))
                        end_action = _ACTIONS["transup"]
                        self.add_to_branch(
                            tau_model, self.current_branch, discon_branch,
                            is_p_wave, end_action)
                    else:
                        # Downgoing section, must look at leg after next to
                        # determine whether to convert on the downgoing or
                        # upgoing part of the path.
                        next_next_leg = self.legs[leg_num + 2]
                        if next_next_leg == "p" or next_next_leg == "s":
                            # Convert on upgoing section
                            end_action = _ACTIONS["turn"]
                            self.add_to_branch(
                                tau_model, self.current_branch,
                                tau_model.cmb_branch - 1, is_p_wave,
                                end_action)
                            end_action = _ACTIONS["transup"]
                            self.add_to_branch(tau_model, self.current_branch,
                                               discon_branch, is_p_wave,
                                               end_action)
                        elif next_next_leg == "P" or next_next_leg == "S":
                            if discon_branch > self.current_branch:
                                # discon is below current loc
                                end_action = _ACTIONS["transdown"]
                                self.add_to_branch(
                                    tau_model, self.current_branch,
                                    discon_branch - 1, is_p_wave, end_action)
                            else:
                                # Discontinuity is above current location,
                                # but we have a downgoing ray, so this is an
                                # illegal ray for this source depth.
                                self.max_ray_param = -1
                                return
                        else:
                            raise TauModelError(
                                "Phase not recognized: {} followed by {} "
                                "followed by {}".format(current_leg, next_leg,
                                                        next_next_leg))
                else:
                    raise TauModelError(
                        "Phase not recognized: {} followed by {}".format(
                            current_leg, next_leg))

            elif current_leg[0] in "PS":
                if current_leg == "Pdiff" or current_leg == "Sdiff":
                    # In the diffracted case we trick addtoBranch into
                    # thinking we are turning, but then make max_ray_param
                    # equal to min_ray_param, which is the deepest turning ray.
                    if (self.max_ray_param >= tau_model.get_tau_branch(
                            tau_model.cmb_branch - 1,
                            is_p_wave).min_turn_ray_param >=
                            self.min_ray_param):
                        if (self.current_branch < tau_model.cmb_branch - 1 or
                            (self.current_branch == tau_model.cmb_branch - 1
                             and end_action != _ACTIONS["diffract"])
                                and prev_leg[0] != "K"):
                            end_action = _ACTIONS["diffract"]
                            self.add_to_branch(
                                tau_model, self.current_branch,
                                tau_model.cmb_branch - 1, is_p_wave,
                                end_action)
                        # otherwise we are already at the right branch to
                        # diffract. remember where the diff or head happened
                        # (one less than size)
                        self.head_or_diffract_seq.append(
                            len(self.branch_seq) - 1)
                        self.max_ray_param = tau_model.get_tau_branch(
                            tau_model.cmb_branch - 1, is_p_wave
                            ).min_turn_ray_param
                        self.min_ray_param = self.max_ray_param

                        if next_leg == "END":
                            end_action = _ACTIONS["reflect_underside"]
                            self.add_to_branch(tau_model, self.current_branch,
                                               upgoing_rec_branch, is_p_wave,
                                               end_action)
                        elif next_leg[0] == "K":
                            end_action = _ACTIONS["transdown"]
                            self.current_branch += 1
                        elif next_leg[0] in "PS":
                            end_action = _ACTIONS["reflect_underside"]
                            self.add_to_branch(
                                tau_model, self.current_branch, 0, is_p_wave,
                                end_action)
                        else:
                            raise TauModelError(
                                "Phase not recognized: "
                                "{} followed by {}".format(
                                    current_leg, next_leg))
                    else:
                        # Can't have head wave as ray param is not within
                        # range.
                        self.max_ray_param = -1
                        return
                elif current_leg in ("Pg", "Sg", "Pn", "Sn"):
                    if self.current_branch >= tau_model.moho_branch:
                        # Pg, Pn, Sg and Sn must be above the moho and so is
                        # not valid for rays coming upwards from below,
                        # possibly due to the source depth. Setting
                        # max_ray_param = -1 effectively disallows this phase.
                        self.max_ray_param = -1
                        return
                    if current_leg in ("Pg", "Sg"):
                        end_action = _ACTIONS["turn"]
                        self.add_to_branch(tau_model, self.current_branch,
                                           tau_model.moho_branch - 1,
                                           is_p_wave, end_action)
                        end_action = _ACTIONS["reflect_underside"]
                        self.add_to_branch(tau_model, self.current_branch,
                                           upgoing_rec_branch, is_p_wave,
                                           end_action)
                    elif current_leg in ("Pn", "Sn"):
                        # In the diffracted case we trick addtoBranch into
                        # thinking we are turning below the Moho, but then
                        # make the min_ray_param equal to max_ray_param,
                        # which is the head wave ray.
                        if (self.max_ray_param >= tau_model.get_tau_branch(
                                tau_model.moho_branch,
                                is_p_wave).max_ray_param >=
                                self.min_ray_param):
                            end_action = _ACTIONS["turn"]
                            self.add_to_branch(
                                tau_model, self.current_branch,
                                tau_model.moho_branch, is_p_wave, end_action)
                            # remember where the head happened
                            self.head_or_diffract_seq.append(
                                len(self.branch_seq) - 1)
                            end_action = _ACTIONS["transup"]
                            self.add_to_branch(
                                tau_model, self.current_branch,
                                tau_model.moho_branch, is_p_wave, end_action)
                            self.min_ray_param = self.max_ray_param
                            if next_leg == "END":
                                end_action = _ACTIONS["reflect_underside"]
                                self.add_to_branch(
                                    tau_model, self.current_branch,
                                    upgoing_rec_branch, is_p_wave, end_action)
                            elif next_leg[0] in "PS":
                                end_action = _ACTIONS["reflect_underside"]
                                self.add_to_branch(
                                    tau_model, self.current_branch, 0,
                                    is_p_wave, end_action)
                        else:
                            # Can't have head wave as ray param is not
                            # within range.
                            self.max_ray_param = -1
                            return
                else:
                    raise TauModelError(
                        "Phase not recognized: {} followed by {}".format(
                            current_leg, next_leg))

            elif current_leg == "K":
                # Now deal with K.
                if tau_model.cmb_depth == tau_model.radius_of_planet:
                    # degenerate case, CMB is at center, so model
                    # without a core
                    self.max_ray_param = -1
                    return
                if tau_model.cmb_depth == tau_model.iocb_depth:
                    # degenerate case, CMB is same as IOCB, so model
                    # without an outer core
                    self.max_ray_param = -1
                    return
                if next_leg in ("P", "S", "p", "s", "Pdiff", "Sdiff"):
                    if prev_leg in ("P", "S", "Pdiff", "Sdiff",
                                    "K", "k", "Kdiff", "START"):
                        end_action = _ACTIONS["turn"]
                        self.add_to_branch(
                            tau_model, self.current_branch,
                            tau_model.iocb_branch - 1, is_p_wave, end_action)
                    end_action = _ACTIONS["transup"]
                    self.add_to_branch(
                        tau_model, self.current_branch, tau_model.cmb_branch,
                        is_p_wave, end_action)
                elif next_leg[0] == "K":
                    if prev_leg in ("P", "S", "K", "Pdiff", "Sdiff", "Kdiff"):
                        end_action = _ACTIONS["turn"]
                        self.add_to_branch(
                            tau_model, self.current_branch,
                            tau_model.iocb_branch - 1, is_p_wave, end_action)
                    end_action = _ACTIONS["reflect_underside"]
                    self.add_to_branch(
                        tau_model, self.current_branch, tau_model.cmb_branch,
                        is_p_wave, end_action)
                elif next_leg in ("I", "J"):
                    # And now consider inner core, I and J.
                    if tau_model.iocb_depth == tau_model.radius_of_planet:
                        # degenerate case, IOCB is at center,
                        # so model without a inner core
                        self.max_ray_param = -1
                        return
                    end_action = _ACTIONS["transdown"]
                    self.add_to_branch(
                        tau_model, self.current_branch,
                        tau_model.iocb_branch - 1, is_p_wave, end_action)
                elif next_leg == "i":
                    end_action = _ACTIONS["reflect_topside"]
                    self.add_to_branch(
                        tau_model, self.current_branch,
                        tau_model.iocb_branch - 1, is_p_wave, end_action)
                else:
                    raise TauModelError(
                        "Phase not recognized: {} followed by {}".format(
                            current_leg, next_leg))

            elif current_leg == "Kdiff":
                # inner core - outer core boundary diffraction
                if (self.max_ray_param >= tau_model.get_tau_branch(
                        tau_model.iocb_branch - 1,
                        is_p_wave).min_turn_ray_param >=
                        self.min_ray_param):
                    if (self.current_branch < tau_model.iocb_branch - 1 or
                        (self.current_branch == tau_model.iocb_branch - 1
                            and end_action != _ACTIONS["diffract"])):
                        end_action = _ACTIONS["diffract"]
                        self.add_to_branch(
                            tau_model, self.current_branch,
                            tau_model.iocb_branch - 1, is_p_wave,
                            end_action)
                    # otherwise we are already at the right branch to
                    # diffract. remember where the diff or head happened
                    # (one less than size)
                    self.head_or_diffract_seq.append(
                        len(self.branch_seq) - 1)
                    self.max_ray_param = tau_model.get_tau_branch(
                        tau_model.iocb_branch - 1, is_p_wave
                        ).min_turn_ray_param
                    self.min_ray_param = self.max_ray_param
                    if next_leg[0] == "K":
                        end_action = _ACTIONS["reflect_underside"]
                        self.add_to_branch(
                            tau_model, self.current_branch,
                            tau_model.cmb_branch,
                            is_p_wave, end_action)
                    if next_leg[0] in ("P", "S"):
                        end_action = _ACTIONS["transup"]
                        self.add_to_branch(
                            tau_model, self.current_branch,
                            tau_model.cmb_branch,
                            is_p_wave, end_action)

                else:
                    raise TauModelError(
                        "Phase not recognized: "
                        "{} followed by {}".format(
                            current_leg, next_leg))

            elif current_leg in ("I", "J"):
                end_action = _ACTIONS["turn"]
                self.add_to_branch(
                    tau_model, self.current_branch,
                    tau_model.tau_branches.shape[1] - 1, is_p_wave, end_action)
                if next_leg in ("I", "J"):
                    end_action = _ACTIONS["reflect_underside"]
                    self.add_to_branch(
                        tau_model, self.current_branch, tau_model.iocb_branch,
                        is_p_wave, end_action)
                elif next_leg[0] == "K":
                    end_action = _ACTIONS["transup"]
                    self.add_to_branch(
                        tau_model, self.current_branch, tau_model.iocb_branch,
                        is_p_wave, end_action)

            elif current_leg in ("m", "c", "i") or current_leg[0] == "^":
                pass

            elif current_leg[0] == "v":
                b = closest_branch_to_depth(tau_model, current_leg[1:])
                if b == 0:
                    raise TauModelError(
                        "Phase not recognized: {} looks like a top side "
                        "reflection at the free surface.".format(current_leg))

            elif is_leg_depth:
                # Check for phase like P0s, but could also be P2s if first
                # discontinuity is deeper.
                b = closest_branch_to_depth(tau_model, current_leg)
                if b == 0 and next_leg in ("p", "s"):
                    raise TauModelError(
                        "Phase not recognized: {} followed by {} looks like "
                        "an upgoing wave from the free surface as closest "
                        "discontinuity to {} is zero depth.".format(
                            current_leg, next_leg, current_leg))

            else:
                raise TauModelError(
                    "Phase not recognized: {} followed by {}".format(
                        current_leg, next_leg))

        if self.max_ray_param != -1:
            if (end_action == _ACTIONS["reflect_underside"] and
                    downgoing_rec_branch == self.branch_seq[-1]):
                # Last action was upgoing, so last branch should be
                # upgoing_rec_branch
                self.min_ray_param = -1
                self.max_ray_param = -1
            elif (end_action == _ACTIONS["reflect_topside"] and
                    upgoing_rec_branch == self.branch_seq[-1]):
                # Last action was downgoing, so last branch should be
                # downgoing_rec_branch
                self.min_ray_param = -1
                self.max_ray_param = -1

    def phase_conversion(self, tau_model, from_branch, end_action, is_p_to_s):
        """
        Change max_ray_param and min_ray_param where there is a phase
        conversion.

        For instance, SKP needs to change the max_ray_param because there are
        SKS ray parameters that cannot propagate from the CMB into the mantle
        as a P wave.
        """
        if end_action == _ACTIONS["turn"]:
            # Can't phase convert for just a turn point
            raise TauModelError("Bad end_action: phase conversion is not "
                                "allowed at turn points.")
        elif end_action == _ACTIONS["reflect_underside"]:
            self.max_ray_param = \
                min(self.max_ray_param,
                    tau_model.get_tau_branch(from_branch,
                                             is_p_to_s).max_ray_param,
                    tau_model.get_tau_branch(from_branch,
                                             not is_p_to_s).max_ray_param)
        elif end_action == _ACTIONS["reflect_topside"]:
            self.max_ray_param = min(
                self.max_ray_param,
                tau_model.get_tau_branch(from_branch,
                                         is_p_to_s).min_turn_ray_param,
                tau_model.get_tau_branch(from_branch,
                                         not is_p_to_s).min_turn_ray_param)
        elif end_action == _ACTIONS["transup"]:
            self.max_ray_param = min(
                self.max_ray_param,
                tau_model.get_tau_branch(from_branch, is_p_to_s).max_ray_param,
                tau_model.get_tau_branch(from_branch - 1,
                                         not is_p_to_s).min_turn_ray_param)
        elif end_action == _ACTIONS["transdown"]:
            self.max_ray_param = min(
                self.max_ray_param,
                tau_model.get_tau_branch(from_branch, is_p_to_s).min_ray_param,
                tau_model.get_tau_branch(from_branch + 1,
                                         not is_p_to_s).max_ray_param)
        else:
            raise TauModelError("Illegal end_action = {}".format(end_action))

    def add_to_branch(self, tau_model, start_branch, end_branch, is_p_wave,
                      end_action):
        """
        Add branch numbers to branch_seq.

        Branches from start_branch to end_branch, inclusive, are added in
        order. Also, current_branch is set correctly based on the value of
        end_action. end_action can be one of transup, transdown,
        reflect_underside, reflect_topside, or turn.
        """
        if end_branch < 0 or end_branch > tau_model.tau_branches.shape[1]:
            raise ValueError('End branch outside range: %d' % (end_branch, ))

        if end_action == _ACTIONS["turn"]:
            end_offset = 0
            is_down_going = True
            self.min_ray_param = max(
                self.min_ray_param,
                tau_model.get_tau_branch(end_branch,
                                         is_p_wave).min_turn_ray_param)
            # careful if the ray param cannot turn due to high slowness.
            # Do not use these layers if their top is in high slowness for the
            # given ray parameter and the bottom is not a critical reflection,
            # rp > max rp in next branch
            b_num = end_branch
            while (b_num >= start_branch):
                branch_top_in_high_slowness = (
                    tau_model.s_mod.depth_in_high_slowness(
                        tau_model.get_tau_branch(b_num, is_p_wave).top_depth,
                        self.min_ray_param,
                        is_p_wave))
                is_bottom_branch = (
                    b_num + 1 >= len(tau_model.tau_branches[0]))
                not_critical_reflect = (
                    is_bottom_branch or
                    self.min_ray_param <=
                    tau_model.get_tau_branch(b_num+1,
                                             is_p_wave
                                             ).max_ray_param)
                if branch_top_in_high_slowness and not_critical_reflect:
                    # tau branch is in high slowness, so turn is not possible,
                    # only non-critical reflect, so do not add these branches
                    end_branch = b_num - 1
                    b_num -= 1
                else:
                    # can turn in b_num layer, so don't worry about shallower
                    # high slowness layers
                    break

        elif end_action == _ACTIONS["reflect_underside"]:
            end_offset = 0
            is_down_going = False
            self.max_ray_param = min(
                self.max_ray_param,
                tau_model.get_tau_branch(end_branch, is_p_wave).max_ray_param)
        elif end_action == _ACTIONS["reflect_topside"]:
            end_offset = 0
            is_down_going = True
            self.max_ray_param = min(
                self.max_ray_param,
                tau_model.get_tau_branch(end_branch,
                                         is_p_wave).min_turn_ray_param)
        elif end_action == _ACTIONS["transup"]:
            end_offset = -1
            is_down_going = False
            self.max_ray_param = min(
                self.max_ray_param,
                tau_model.get_tau_branch(end_branch, is_p_wave).max_ray_param)
        elif end_action == _ACTIONS["transdown"]:
            end_offset = 1
            is_down_going = True
            self.max_ray_param = min(
                self.max_ray_param,
                tau_model.get_tau_branch(end_branch, is_p_wave).min_ray_param)
        elif end_action == _ACTIONS["diffract"]:
            end_offset = 0
            is_down_going = True
            self.max_ray_param = min(
                self.max_ray_param,
                tau_model.get_tau_branch(
                    end_branch, is_p_wave).min_turn_ray_param)
        else:
            raise TauModelError("Illegal end_action: {}".format(end_action))

        segment = SeismicPhaseSegment(self.tau_model, start_branch, end_branch,
                                      is_p_wave, end_action, is_down_going)
        if self.segment_list:
            # Perform some sanity checks on the segment being added
            prev_segment = self.segment_list[-1]
            _check_segment_consistency(prev_segment, segment)
        self.segment_list.append(segment)

        if is_down_going:
            if start_branch > end_branch:
                # Can't be downgoing as we are already below.
                self.min_ray_param = -1
                self.max_ray_param = -1
            else:
                # Must be downgoing, so increment i.
                for i in range(start_branch, end_branch + 1):
                    self.branch_seq.append(i)
                    self.down_going.append(is_down_going)
                    self.wave_type.append(is_p_wave)
        else:
            if start_branch < end_branch:
                # Can't be upgoing as we are already above.
                self.min_ray_param = -1
                self.max_ray_param = -1
            else:
                # Upgoing, so decrement i.
                for i in range(start_branch, end_branch - 1, -1):
                    self.branch_seq.append(i)
                    self.down_going.append(is_down_going)
                    self.wave_type.append(is_p_wave)
        self.current_branch = end_branch + end_offset

    def sum_branches(self, tau_model):
        """Sum the appropriate branches for this phase."""
        # Special case for surface waves.
        if self.name.endswith("kmps"):
            self.dist = np.zeros(2)
            self.time = np.zeros(2)
            self.ray_param = np.empty(2, dtype=np.float64)

            self.ray_param[0] = \
                tau_model.radius_of_planet / float(self.name[:-4])

            self.dist[1] = 2 * math.pi
            self.time[1] = \
                2 * math.pi * tau_model.radius_of_planet / \
                float(self.name[:-4])
            self.ray_param[1] = self.ray_param[0]

            self.min_distance = 0
            self.max_distance = 2 * math.pi
            return

        if self.max_ray_param < 0 or self.min_ray_param > self.max_ray_param:
            # Phase has no arrivals, possibly due to source depth.
            self.ray_param = np.empty(0, dtype=np.float64)
            self.min_ray_param = -1
            self.max_ray_param = -1
            self.dist = np.empty(0)
            self.time = np.empty(0)
            self.max_distance = -1
            return

        # Find the ray parameter index that corresponds to the min_ray_param
        # and max_ray_param.
        index = np.where(tau_model.ray_params >= self.min_ray_param)[0]
        if len(index):
            self.min_ray_param_index = index[-1]
        index = np.where(tau_model.ray_params >= self.max_ray_param)[0]
        if len(index):
            self.max_ray_param_index = index[-1]
        if self.max_ray_param_index == 0 \
                and self.min_ray_param_index == len(tau_model.ray_params) - 1:
            # All ray parameters are valid so just copy:
            self.ray_param = tau_model.ray_params.copy()
        elif self.max_ray_param_index == self.min_ray_param_index:
            # if "Sdiff" in self.name or "Pdiff" in self.name:
            # self.ray_param = [self.min_ray_param, self.min_ray_param]
            # elif "Pn" in self.name or "Sn" in self.name:
            # self.ray_param = [self.min_ray_param, self.min_ray_param]
            if self.name.endswith("kmps"):
                self.ray_param = np.array([0.0, self.max_ray_param],
                                          dtype=np.float64)
            else:
                self.ray_param = np.array([self.min_ray_param,
                                           self.min_ray_param],
                                          dtype=np.float64)
        else:
            # Only a subset of the ray parameters is valid so use these.
            self.ray_param = \
                tau_model.ray_params[self.max_ray_param_index:
                                     self.min_ray_param_index + 1].copy()

        self.dist = np.zeros(shape=self.ray_param.shape)
        self.time = np.zeros(shape=self.ray_param.shape)

        # Counter for passes through each branch. 0 is P and 1 is S.
        times_branches = self.calc_branch_mult(tau_model)

        # Sum the branches with the appropriate multiplier.
        size = self.min_ray_param_index - self.max_ray_param_index + 1
        index = slice(self.max_ray_param_index, self.min_ray_param_index + 1)
        for i in range(tau_model.tau_branches.shape[1]):
            tb = times_branches[0, i]
            tbs = times_branches[1, i]
            taub = tau_model.tau_branches[0, i]
            taubs = tau_model.tau_branches[1, i]

            if tb != 0:
                self.dist[:size] += tb * taub.dist[index]
                self.time[:size] += tb * taub.time[index]
            if tbs != 0:
                self.dist[:size] += tbs * taubs.dist[index]
                self.time[:size] += tbs * taubs.time[index]

        if "Sdiff" in self.name or "Pdiff" in self.name:
            if (tau_model.cmb_depth == tau_model.radius_of_planet
                or tau_model.s_mod.depth_in_high_slowness(
                    tau_model.cmb_depth - 1e-10, self.min_ray_param,
                    self.name[0] == "P")):
                # No diffraction if cmb is zero radius or
                # there is a high slowness zone at the CMB.
                self.min_ray_param = -1
                self.max_ray_param = -1
                self.max_distance = -1
                self.time = np.empty(0)
                self.dist = np.empty(0)
                self.ray_param = np.empty(0)
                return
            else:
                self.dist[1] = self.dist[0] + \
                    self._settings["max_diffraction_in_radians"]
                self.time[1] = (
                    self.time[0] +
                    self._settings["max_diffraction_in_radians"] *
                    self.min_ray_param)

        elif "Kdiff" in self.name:
            self.dist[1] = self.dist[0] + \
                self._settings["max_diffraction_in_radians"]
            self.time[1] = (
                self.time[0] +
                self._settings["max_diffraction_in_radians"] *
                self.min_ray_param)

        elif "Pn" in self.name or "Sn" in self.name:
            self.dist[1] = self.dist[0] + \
                self._settings["max_refraction_in_radians"]
            self.time[1] = (
                self.time[0] + self._settings["max_refraction_in_radians"] *
                self.min_ray_param)

        elif self.max_ray_param_index == self.min_ray_param_index:
            self.dist[1] = self.dist[0]
            self.time[1] = self.time[0]

        self.min_distance = np.min(self.dist)
        self.max_distance = np.max(self.dist)

        # Now check to see if our ray parameter range includes any ray
        # parameters that are associated with high slowness zones. If so,
        # then we will need to insert a "shadow zone" into our time and
        # distance arrays. It is represented by a repeated ray parameter.
        for is_pwave in [True, False]:
            hsz = tau_model.s_mod.high_slowness_layer_depths_p \
                if is_pwave \
                else tau_model.s_mod.high_slowness_layer_depths_s
            index_offset = 0
            for hszi in hsz:
                if self.max_ray_param > hszi.ray_param > self.min_ray_param:
                    # There is a high slowness zone within our ray parameter
                    # range so might need to add a shadow zone. Need to
                    # check if the current wave type is part of the phase at
                    # this depth/ray parameter.
                    branch_num = tau_model.find_branch(hszi.top_depth)
                    found_overlap = False
                    for leg_num in range(len(self.branch_seq)):
                        # Check for downgoing legs that cross the high
                        # slowness zone with the same wave type.
                        if (self.branch_seq[leg_num] == branch_num and
                                self.wave_type[leg_num] == is_pwave and
                                self.down_going[leg_num] is True and
                                self.branch_seq[leg_num - 1] ==
                                branch_num - 1 and
                                self.wave_type[leg_num - 1] == is_pwave and
                                self.down_going[leg_num - 1] is True):
                            found_overlap = True
                            break
                    if found_overlap:
                        hsz_index = np.where(self.ray_param == hszi.ray_param)
                        hsz_index = hsz_index[0][0]

                        newlen = self.ray_param.shape[0] + 1
                        new_ray_params = np.empty(newlen)
                        newdist = np.empty(newlen)
                        newtime = np.empty(newlen)

                        new_ray_params[:hsz_index] = self.ray_param[:hsz_index]
                        newdist[:hsz_index] = self.dist[:hsz_index]
                        newtime[:hsz_index] = self.time[:hsz_index]

                        # Sum the branches with an appropriate multiplier.
                        new_ray_params[hsz_index] = hszi.ray_param
                        newdist[hsz_index] = 0
                        newtime[hsz_index] = 0
                        for tb, tbs, taub, taubs in zip(
                                times_branches[0], times_branches[1],
                                tau_model.tau_branches[0],
                                tau_model.tau_branches[1]):
                            if tb != 0 and taub.top_depth < hszi.top_depth:
                                newdist[hsz_index] += tb * taub.dist[
                                    self.max_ray_param_index + hsz_index -
                                    index_offset]
                                newtime[hsz_index] += tb * taub.time[
                                    self.max_ray_param_index + hsz_index -
                                    index_offset]
                            if tbs != 0 and taubs.top_depth < hszi.top_depth:
                                newdist[hsz_index] += tbs * taubs.dist[
                                    self.max_ray_param_index + hsz_index -
                                    index_offset]
                                newtime[hsz_index] += tbs * taubs.time[
                                    self.max_ray_param_index + hsz_index -
                                    index_offset]

                        newdist[hsz_index + 1:] = self.dist[hsz_index:]
                        newtime[hsz_index + 1:] = self.time[hsz_index:]
                        new_ray_params[hsz_index + 1:] = \
                            self.ray_param[hsz_index:]

                        index_offset += 1
                        self.dist = newdist
                        self.time = newtime
                        self.ray_param = new_ray_params

    def calc_branch_mult(self, tau_model):
        """
        Calculate how many times the phase passes through a branch, up or down.

        With this result, we can just multiply instead of doing the ray calc
        for each time.
        """
        # Initialise the counter for each branch to 0. 0 is P and 1 is S.
        times_branches = np.zeros((2, tau_model.tau_branches.shape[1]))
        # Count how many times each branch appears in the path.
        # wave_type is at least as long as branch_seq
        for wt, bs in zip(self.wave_type, self.branch_seq):
            if wt:
                times_branches[0][bs] += 1
            else:
                times_branches[1][bs] += 1
        return times_branches

    def calc_time(self, degrees,
                  ray_param_tol=_DEFAULT_VALUES["default_time_ray_param_tol"]):
        """
        Calculate arrival times for this phase, sorted by time.
        """
        # 100 should finally be enough...this in only for one phase after
        # all..
        r_dist = np.empty(100, dtype=np.float64)
        r_ray_num = np.empty(100, dtype=np.int32)

        # This saves around 17% runtime when calculating arrival times which
        # is probably the major use case.
        phase_count = clibtau.seismic_phase_calc_time_inner_loop(
            float(degrees),
            self.max_distance,
            self.dist,
            self.ray_param,
            r_dist,
            r_ray_num,
            len(self.dist)
        )

        arrivals = []
        for _i in range(phase_count):
            arrivals.append(self.refine_arrival(
                degrees, r_ray_num[_i], r_dist[_i], ray_param_tol,
                self._settings["max_recursion"]))
        return arrivals

    def calc_pierce(self, degrees,
                    ray_param_tol=_DEFAULT_VALUES["default_path_ray_param_tol"]
                    ):
        """
        Calculate pierce points for this phase.

        First calculates arrivals, then the "pierce points" corresponding to
        the stored arrivals. The pierce points are stored within each arrival
        object.
        """
        arrivals = self.calc_time(degrees, ray_param_tol=ray_param_tol)
        for arrival in arrivals:
            self.calc_pierce_from_arrival(arrival)
        return arrivals

    def calc_pierce_from_arrival(self, curr_arrival):
        """
        Calculate the pierce points for a particular arrival.

        The returned arrival is the same as the input argument but now has the
        pierce points filled in.
        """
        # Find the ray parameter index that corresponds to the arrival ray
        # parameter in the TauModel, ie it is between ray_num and ray_num+1,
        # We know that it must be <model.ray_param.length-1 since the last
        # ray parameter sample is 0 in a spherical model.
        ray_num = 0
        for i, rp in enumerate(self.tau_model.ray_params[:-1]):
            if rp >= curr_arrival.ray_param:
                ray_num = i
            else:
                break

        # Here we use ray parameter and dist info stored within the
        # SeismicPhase so we can use curr_arrival.ray_param_index, which
        # may not correspond to ray_num (for model.ray_param).
        ray_param_a = self.ray_param[curr_arrival.ray_param_index]
        ray_param_b = self.ray_param[curr_arrival.ray_param_index + 1]
        dist_a = self.dist[curr_arrival.ray_param_index]
        dist_b = self.dist[curr_arrival.ray_param_index + 1]
        dist_ratio = (curr_arrival.purist_dist - dist_a) / (dist_b - dist_a)
        dist_ray_param = dist_ratio * (ray_param_b - ray_param_a) + ray_param_a

        # + 2 for first point and kmps, if it exists.
        pierce = np.empty(len(self.branch_seq) +
                          len(self.head_or_diffract_seq) + 2, dtype=TimeDist)
        # First pierce point is always 0 distance at the source depth.
        pierce[0] = (dist_ray_param, 0, 0, self.tau_model.source_depth)
        index = 1
        branch_dist = 0
        branch_time = 0

        # Loop from 0 but already done 0 [I just copy the comments, sorry!],
        # so the pierce point when the ray leaves branch i is stored in i + 1.
        # Use linear interpolation between rays that we know.
        assert len(self.branch_seq) == len(self.wave_type) == \
            len(self.down_going)
        for i, branch_num, is_p_wave, is_down_going in zip(
                count(), self.branch_seq, self.wave_type, self.down_going):
            # Save the turning depths for the ray parameter for both P and
            # S waves. This way we get the depth correct for any rays that
            # turn within a layer. We have to do this on a per branch basis
            # because of converted phases, e.g. SKS.
            tau_branch = self.tau_model.get_tau_branch(branch_num, is_p_wave)
            if dist_ray_param > tau_branch.max_ray_param:
                turn_depth = tau_branch.top_depth
            elif dist_ray_param <= tau_branch.min_ray_param:
                turn_depth = tau_branch.bot_depth
            else:
                if (is_p_wave or self.tau_model.s_mod.depth_in_fluid((
                        tau_branch.top_depth + tau_branch.bot_depth) / 2)):
                    turn_depth = self.tau_model.s_mod.find_depth_from_depths(
                        dist_ray_param,
                        tau_branch.top_depth,
                        tau_branch.bot_depth,
                        True)
                else:
                    turn_depth = self.tau_model.s_mod.find_depth_from_depths(
                        dist_ray_param,
                        tau_branch.top_depth,
                        tau_branch.bot_depth,
                        is_p_wave)

            if self.head_or_diffract_seq:
                # Head waves and diffracted waves are a special case.
                dist_a = tau_branch.dist[ray_num]
                time_a = tau_branch.time[ray_num]
                dist_b, time_b = dist_a, time_a
            else:
                dist_a = tau_branch.dist[ray_num]
                time_a = tau_branch.time[ray_num]
                dist_b = tau_branch.dist[ray_num + 1]
                time_b = tau_branch.time[ray_num + 1]

            branch_dist += dist_ratio * (dist_b - dist_a) + dist_a
            prev_branch_time = np.array(branch_time, copy=True)
            branch_time += dist_ratio * (time_b - time_a) + time_a
            if is_down_going:
                branch_depth = min(tau_branch.bot_depth, turn_depth)
            else:
                branch_depth = min(tau_branch.top_depth, turn_depth)

            # Make sure ray actually propagates in this branch; leave a little
            # room for numerical chatter.
            if abs(prev_branch_time - branch_time) > 1e-10:
                pierce[index] = (dist_ray_param, branch_time, branch_dist,
                                 branch_depth)
                index += 1

            if i in self.head_or_diffract_seq:
                refract_dist = (
                    curr_arrival.purist_dist - self.dist[0]
                    )/len(self.head_or_diffract_seq)
                refract_time = refract_dist * curr_arrival.ray_param
                pierce[index] = (dist_ray_param,
                                 branch_time + refract_time,
                                 branch_dist + refract_dist,
                                 branch_depth)
                branch_dist += refract_dist
                prev_branch_time = branch_time
                branch_time += refract_time
                index += 1

        if "kmps" in self.name:
            pierce[index] = (dist_ray_param, curr_arrival.time,
                             curr_arrival.purist_dist, 0)
            index += 1

        curr_arrival.pierce = pierce[:index]
        # The arrival is modified in place and must (?) thus be returned.
        return curr_arrival

    def calc_path(self, degrees,
                  ray_param_tol=_DEFAULT_VALUES["default_path_ray_param_tol"]):
        """
        Calculate the paths this phase takes through the planet model.

        Only calls :meth:`calc_path_from_arrival`.
        """
        arrivals = self.calc_time(degrees, ray_param_tol=ray_param_tol)
        for arrival in arrivals:
            self.calc_path_from_arrival(arrival)
        return arrivals

    def calc_path_from_arrival(self, curr_arrival):
        """
        Calculate the paths this phase takes through the planet model.
        """
        # Find the ray parameter index that corresponds to the arrival ray
        # parameter in the TauModel, i.e. it is between ray_num and
        # ray_num + 1.
        temp_time_dist = (curr_arrival.ray_param, 0, 0,
                          self.tau_model.source_depth)
        # path_list is a list of lists.
        path_list = [temp_time_dist]
        for i, branch_num, is_p_wave, is_down_going in zip(
                count(), self.branch_seq, self.wave_type, self.down_going):
            br = self.tau_model.get_tau_branch(branch_num, is_p_wave)
            temp_time_dist = br.path(curr_arrival.ray_param, is_down_going,
                                     self.tau_model.s_mod)
            if len(temp_time_dist):
                path_list.extend(temp_time_dist)
                if np.any(temp_time_dist['dist'] < 0):
                    raise RuntimeError("Path is backtracking, "
                                       "this is impossible.")

            # Special case for head and diffracted waves:
            if i in self.head_or_diffract_seq:
                dist_diff = (
                    curr_arrival.purist_dist - self.dist[0]
                    )/len(self.head_or_diffract_seq)
                if "Pdiff" in self.name or "Sdiff" in self.name:
                    branch_depth = self.tau_model.cmb_depth
                elif "Pn" in self.name or "Sn" in self.name:
                    branch_depth = self.tau_model.moho_depth
                elif "Kdiff" in self.name:
                    branch_depth = self.tau_model.iocb_depth
                else:
                    raise RuntimeError(
                        "Path adding head and diffracted wave, "
                        "but did not find P/Sdiff or P/Sn, expected")
                diff_td = (
                    curr_arrival.ray_param,
                    dist_diff * curr_arrival.ray_param,
                    dist_diff,
                    branch_depth)

                path_list.append(diff_td)

        if "kmps" in self.name:
            # kmps phases have no branches, so need to end them at the arrival
            # distance.
            head_td = (
                curr_arrival.ray_param,
                curr_arrival.purist_dist * curr_arrival.ray_param,
                curr_arrival.purist_dist,
                0)
            path_list.append(head_td)

        curr_arrival.path = np.array(path_list, dtype=TimeDist)
        np.cumsum(curr_arrival.path['time'], out=curr_arrival.path['time'])
        np.cumsum(curr_arrival.path['dist'], out=curr_arrival.path['dist'])

        return curr_arrival

    def refine_arrival(self, degrees, ray_index, dist_radian, tolerance,
                       recursion_limit):
        """
        Use a shooting method to improve ray path.
        """

        left = Arrival(self, degrees, self.time[ray_index],
                       self.dist[ray_index], self.ray_param[ray_index],
                       ray_index, self.name, self.purist_name,
                       self.source_depth, self.receiver_depth)
        right = Arrival(self, degrees, self.time[ray_index + 1],
                        self.dist[ray_index + 1],
                        self.ray_param[ray_index + 1],
                        # Use ray_index since dist is between ray_index and
                        # (ray_index + 1).
                        ray_index, self.name, self.purist_name,
                        self.source_depth, self.receiver_depth)

        new_estimate = self.linear_interp_arrival(degrees, dist_radian,
                                                  left, right)

        # can't shoot/refine for non-body waves
        if (recursion_limit <= 0 or self.name.endswith('kmps') or
                self.head_or_diffract_seq):
            return new_estimate

        # Find more accurate ray parameter by root-finding
        new_arrivals = [new_estimate]
        left_ray_param = self.ray_param[ray_index]
        right_ray_param = self.ray_param[ray_index + 1]
        left_dist = self.dist[ray_index]
        right_dist = self.dist[ray_index + 1]

        def residual(ray_param):
            if ray_param == left_ray_param:
                dist = left_dist
            elif ray_param == right_ray_param:
                dist = right_dist
            else:
                shoot = self.shoot_ray(degrees, ray_param)
                new_arrivals.append(shoot)
                dist = shoot.purist_dist
            return dist_radian - dist

        new_ray_param = brentq(residual, left_ray_param, right_ray_param,
                               xtol=tolerance, maxiter=recursion_limit,
                               disp=False)
        new_ray_param = np.float64(new_ray_param)

        # the arrival brentq calculated at its last iteration
        last = new_arrivals[-1]

        # Use stationarity of the theta function to get better estimate of time
        # Buland and Chapman (1983), equations (16) and (20)
        theta = last.time + last.ray_param * (dist_radian - last.purist_dist)

        return Arrival(self, degrees, theta,
                       dist_radian, new_ray_param,
                       ray_index, self.name, self.purist_name,
                       self.source_depth, self.receiver_depth)

    def shoot_ray(self, degrees, ray_param):
        if (self.head_or_diffract_seq or
                self.name.endswith('kmps')):
            raise SlownessModelError('Unable to shoot ray in non-body waves')

        if ray_param < self.min_ray_param or self.max_ray_param < ray_param:
            msg = 'Ray param %f is outside range for this phase: min=%f max=%f'
            raise SlownessModelError(msg % (ray_param, self.min_ray_param,
                                            self.max_ray_param))

        # looks like a body wave and ray param can propagate
        for ray_param_index in range(len(self.ray_param) - 1):
            if self.ray_param[ray_param_index + 1] < ray_param:
                break

        tau_model = self.tau_model
        s_mod = tau_model.s_mod

        # counter for passes through each branch. 0 is P and 1 is S.
        times_branches = self.calc_branch_mult(tau_model)
        time = np.zeros(1)
        dist = np.zeros(1)
        ray_param = np.array([ray_param])

        # Sum the branches with the appropriate multiplier.
        for j in range(tau_model.tau_branches.shape[1]):
            if times_branches[0, j] != 0:
                br = tau_model.get_tau_branch(j, s_mod.p_wave)
                top_layer = s_mod.layer_number_below(br.top_depth,
                                                     s_mod.p_wave)
                bot_layer = s_mod.layer_number_above(br.bot_depth,
                                                     s_mod.p_wave)
                td = br.calc_time_dist(s_mod, top_layer, bot_layer, ray_param,
                                       allow_turn_in_layer=True)

                time += times_branches[0, j] * td['time']
                dist += times_branches[0, j] * td['dist']

            if times_branches[1, j] != 0:
                br = tau_model.get_tau_branch(j, s_mod.s_wave)
                top_layer = s_mod.layer_number_below(br.top_depth,
                                                     s_mod.s_wave)
                bot_layer = s_mod.layer_number_above(br.bot_depth,
                                                     s_mod.s_wave)
                td = br.calc_time_dist(s_mod, top_layer, bot_layer, ray_param,
                                       allow_turn_in_layer=True)

                time += times_branches[1, j] * td['time']
                dist += times_branches[1, j] * td['dist']

        return Arrival(self, degrees, time[0], dist[0], ray_param[0],
                       ray_param_index, self.name, self.purist_name,
                       self.source_depth, self.receiver_depth)

    def linear_interp_arrival(self, degrees, search_dist, left, right):
        if left.purist_dist == search_dist:
            return left
        if right.purist_dist == search_dist:
            return right

        dp_dx = (left.ray_param - right.ray_param) / (
                 left.purist_dist - right.purist_dist)
        ray_param = (search_dist - right.purist_dist) * dp_dx + right.ray_param

        # Use stationarity of the theta function to estimate arrival time
        # Buland and Chapman (1983), equations (16) and (20)
        left_theta = left.time + left.ray_param * (
            search_dist - left.purist_dist)
        right_theta = right.time + right.ray_param * (
            search_dist - right.purist_dist)
        if dp_dx > 0:
            arrival_time = max(left_theta, right_theta)
        else:
            arrival_time = min(left_theta, right_theta)

        if math.isnan(arrival_time):
            msg = ('Time is NaN, search=%f leftDist=%f leftTime=%f '
                   'rightDist=%f rightTime=%f')
            raise RuntimeError(msg % (search_dist, left.purist_dist, left.time,
                                      right.purist_dist, right.time))
        return Arrival(self, degrees, arrival_time, search_dist, ray_param,
                       left.ray_param_index, self.name, self.purist_name,
                       self.source_depth, self.receiver_depth)

    def calc_ray_param_for_takeoff(self, takeoff_degree):
        v_mod = self.tau_model.s_mod.v_mod
        try:
            if self.down_going[0]:
                takeoff_velocity = v_mod.evaluate_below(self.source_depth,
                                                        self.name[0])
            else:
                takeoff_velocity = v_mod.evaluate_above(self.source_depth,
                                                        self.name[0])
        except (IndexError, LookupError) as e:
            msg = 'Please contact the developers. This error should not occur.'
            raise RuntimeError(msg) from e

        return ((self.tau_model.radius_of_planet - self.source_depth) *
                math.sin(np.radians(takeoff_degree)) / takeoff_velocity)

    def calc_takeoff_angle(self, ray_param):
        if self.name.endswith('kmps'):
            return 0

        v_mod = self.tau_model.s_mod.v_mod
        try:
            if self.down_going[0]:
                takeoff_velocity = v_mod.evaluate_below(self.source_depth,
                                                        self.name[0])
            else:
                takeoff_velocity = v_mod.evaluate_above(self.source_depth,
                                                        self.name[0])
        except (IndexError, LookupError) as e:
            msg = 'Please contact the developers. This error should not occur.'
            raise RuntimeError(msg) from e

        takeoff_angle = np.degrees(math.asin(np.clip(
            takeoff_velocity.item() * ray_param /
            (self.tau_model.radius_of_planet - self.source_depth), -1.0, 1.0)))
        if not self.down_going[0]:
            # upgoing, so angle is in 90-180 range
            takeoff_angle = 180 - takeoff_angle

        return takeoff_angle

    def calc_incident_angle(self, ray_param):
        if self.name.endswith('kmps'):
            return 0

        v_mod = self.tau_model.s_mod.v_mod
        # Very last item is "END", assume first char is P or S
        last_leg = self.legs[-2][0]
        try:
            if self.down_going[-1]:
                incident_velocity = v_mod.evaluate_above(self.receiver_depth,
                                                         last_leg)
            else:
                incident_velocity = v_mod.evaluate_below(self.receiver_depth,
                                                         last_leg)
        except (IndexError, LookupError) as e:
            msg = 'Please contact the developers. This error should not occur.'
            raise RuntimeError(msg) from e

        incident_angle = np.degrees(math.asin(np.clip(
            incident_velocity.item() * ray_param /
            (self.tau_model.radius_of_planet - self.receiver_depth),
            -1.0, 1.0)))
        if self.down_going[-1]:
            incident_angle = 180 - incident_angle

        return incident_angle

    @classmethod
    def get_earliest_arrival(cls, rel_phases, degrees):
        raise NotImplementedError("baaa")

    def _shadow_zone_splits(self):
        """
        A list of slices that split on any shadow zones.
        """
        if self.head_or_diffract_seq:
            # don't split any diffracted waves
            return [slice(0, None)]
        else:
            # shadow zones are where ray parameter repeats
            return _repeated_value_splits(self.ray_param)


def _repeated_value_splits(x):
    """
    A list of slices that split a 1D numpy array where values repeat.
    """
    split_indices = []
    idxs = np.where(np.diff(x) == 0)[0] + 1
    left_index = 0
    for i in idxs:
        split_indices.append(slice(left_index, i))
        left_index = i
    split_indices.append(slice(left_index, None))

    return split_indices


def closest_branch_to_depth(tau_model, depth_string):
    """
    Find the closest discontinuity to the given depth that can have
    reflections and phase transformations.
    """
    if depth_string == "m":
        return tau_model.moho_branch
    elif depth_string == "c":
        return tau_model.cmb_branch
    elif depth_string == "i":
        return tau_model.iocb_branch
    # Non-standard boundary, given by a number: must look for it.
    discon_branch = -1
    discon_max = 1e300
    discon_depth = float(depth_string)
    for i, t_branch in enumerate(tau_model.tau_branches[0]):
        if (abs(discon_depth - t_branch.top_depth) < discon_max and not
                any(ndc == t_branch.top_depth
                    for ndc in tau_model.no_discon_depths)):
            discon_branch = i
            discon_max = abs(discon_depth - t_branch.top_depth)
    return discon_branch


def self_tokenizer(scanner, token):
    return token


def wrong_phase(scanner, token):
    raise ValueError("Invalid phase name: %s cannot be followed by %s in %s" %
                     (token[0], token[1], token))


tokenizer = re.Scanner([
    # Surface wave phase velocity "phases"
    (r"\.?\d+\.?\d*kmps", self_tokenizer),
    # Composite legs.
    (r"Pn|Sn|Pg|Sg|Pb|Sb|Pdiff|Sdiff|Kdiff|Ped|Sed", self_tokenizer),
    # Reflections.
    (r"([\^v])([mci]|\.?\d+\.?\d*)", self_tokenizer),
    # Invalid phases.
    (r"[PS][ps]", wrong_phase),
    # Single legs.
    (r"[KkIiJmcPpSs]", self_tokenizer),
    # Single numerical value
    (r"\.?\d+\.?\d*", self_tokenizer)
])


def leg_puller(name):
    """
    Tokenize a phase name into legs.

    For example, ``PcS`` becomes ``'P' + 'c' + 'S'`` while ``p^410P`` would
    become ``'p' + '^410' + 'P'``. Once a phase name has been broken into
    tokens, we can begin to construct the sequence of branches to which it
    corresponds. Only minor error checking is done at this point, for instance
    ``PIP`` generates an exception but ``^410`` doesn't. It also appends
    ``"END"`` as the last leg.
    """
    results, remainder = tokenizer.scan(name)
    if remainder:
        raise ValueError("Invalid phase name: %s could not be parsed in %s"
                         % (str(remainder), name))
    return results + ["END"]


class SeismicPhaseSegment(object):
    def __init__(self, tau_model, start_branch, end_branch, is_p_wave,
                 end_action, is_down_going):
        self.tau_model = tau_model
        self.start_branch = start_branch
        self.end_branch = end_branch
        self.is_p_wave = is_p_wave
        self.end_action = end_action
        self.is_down_going = is_down_going


def _check_segment_consistency(prev_segment, segment):
    """
    Sanity checks to see if the SeismicPhaseSegment
    being added is consistent with the previous segment.
    Raises a TauModelError is there is a problem.
    """
    if segment.is_down_going:
        if prev_segment.end_branch > segment.start_branch:
            raise TauModelError(
                f"Segment is downgoing, but we are already below the "
                f"start: {segment.end_action}")
        if prev_segment.end_action == _ACTIONS["reflect_topside"]:
            raise TauModelError(
                f"Segment is downgoing, but previous action was to "
                f"reflect up: {segment.end_action}")
        if prev_segment.end_action == _ACTIONS["turn"]:
            raise TauModelError(
                f"Segment is downgoing, but previous action was to "
                f"turn: {segment.end_action}")
        if prev_segment.end_action == _ACTIONS["transup"]:
            raise TauModelError(
                f"Segment is downgoing, but previous action was to "
                f"transmit up: {segment.end_action}")
        if (prev_segment.end_branch == segment.start_branch
                and not prev_segment.is_down_going and
                prev_segment.end_action !=
                _ACTIONS["reflect_underside"]):
            raise TauModelError(
                f"Segment is downgoing, but previous action was not to"
                f"reflect underside: {segment.end_action}")
    else:
        if prev_segment.end_branch < segment.start_branch:
            raise TauModelError(
                f"Segment is upgoing, but we are already above the "
                f"start: {segment.end_action}")
        if prev_segment.end_action == _ACTIONS["reflect_underside"]:
            raise TauModelError(
                f"Segment is upgoing, but previous action was to "
                f"underside reflect down:  {segment.end_action}")
        if prev_segment.end_action == _ACTIONS["transdown"]:
            raise TauModelError(
                f"Segment is upgoing, but previous action was to "
                f"trans down: {segment.end_action}")
        if (prev_segment.end_branch == segment.start_branch
                and prev_segment.is_down_going and
                prev_segment.end_action not in
                (_ACTIONS["turn"], _ACTIONS["diffract"],
                    _ACTIONS["reflect_topside"])):
            raise TauModelError(
                f"Segment is upgoing, but previous action was not to "
                f"reflect topside: {segment.end_action}")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
import collections
from unittest import mock

from packaging.version import parse as parse_version
import pytest

import obspy
from obspy.clients.fdsn.header import FDSNNoDataException
from obspy.clients.fdsn.routing.eidaws_routing_client import \
    EIDAWSRoutingClient


_DummyResponse = collections.namedtuple("_DummyResponse", ["content"])


@pytest.mark.network
class TestEIDAWSRoutingClient():
    @classmethod
    def setup_class(cls):
        cls.client = EIDAWSRoutingClient(timeout=20)
        cls._cls = ("obspy.clients.fdsn.routing.eidaws_routing_client."
                    "EIDAWSRoutingClient")

    def test_get_service_version(self):
        # At the time of test writing the version is 1.1.1. Here we just
        # make sure it is larger.
        assert parse_version(self.client.get_service_version()) >= \
            parse_version("1.1.1")

    def test_error_handling(self):
        msg = "No data available for request."
        with pytest.raises(FDSNNoDataException, match=msg):
            self.client.get_waveforms(
                network="XX", station="XXXXX", location="XX", channel="XXX",
                starttime=obspy.UTCDateTime(2017, 1, 1),
                endtime=obspy.UTCDateTime(2017, 1, 2))

    def test_get_waveforms_integration_test(self):
        """
        Integration test that does not mock anything but actually downloads
        things.
        """
        st = self.client.get_waveforms(
            network="B*", station="*", location="*", channel="HHZ",
            starttime=obspy.UTCDateTime(2017, 1, 1),
            endtime=obspy.UTCDateTime(2017, 1, 1, 0, 0, 0, 1))
        # This yields 4 channels at the time of writing this test - I assume
        # it is unlikely to every yield less than 1.
        # So this test should be fairly stable.
        assert len(st) >= 1

        # Same with the bulk download.
        st2 = self.client.get_waveforms_bulk(
            [["B*", "*", "*", "HHZ", obspy.UTCDateTime(2017, 1, 1),
              obspy.UTCDateTime(2017, 1, 1, 0, 0, 0, 1)]])
        assert len(st2) >= 1

        # They should be identical.
        assert st == st2

    def test_get_stations_integration_test(self):
        """
        Integration test that does not mock anything but actually downloads
        things.
        """
        inv = self.client.get_stations(
            network="B*", station="*", location="*", channel="LHZ",
            starttime=obspy.UTCDateTime(2017, 1, 1),
            endtime=obspy.UTCDateTime(2017, 1, 1, 0, 1),
            level="network")
        # This yields 1 network at the time of writing this test - I assume
        # it is unlikely to every yield less. So this test should be fairly
        # stable.
        assert len(inv) >= 1

        # Can also be formulated as a bulk query.
        inv2 = self.client.get_stations_bulk(
            [["B*", "*", "*", "LHZ", obspy.UTCDateTime(2017, 1, 1),
              obspy.UTCDateTime(2017, 1, 1, 0, 1)]],
            level="network")
        assert len(inv2) >= 1

        # The results should be basically identical - they will still differ
        # because times stamps and also order might change slightly.
        # But the get_contents() method should be safe enough.
        assert inv.get_contents() == inv2.get_contents()

    def test_proper_no_data_exception_on_out_of_epoch_dates(self):
        """
        Test for #2611 (EIDA/userfeedback#56) which was leading to bulk request
        of *all* EIDA data when querying a legit station but an out-of-epoch
        time window for which no data exists.
        """
        # this time window is before the requested station was installed
        t1 = obspy.UTCDateTime('2012-01-01')
        t2 = t1 + 2
        with pytest.raises(FDSNNoDataException, match='No data'):
            self.client.get_waveforms(
                network='OE', station='UNNA', channel='HHZ', location='*',
                starttime=t1, endtime=t2)


class TestEIDAWSRoutingClientNoNetwork():
    @classmethod
    def setup_class(cls):
        cls.client = EIDAWSRoutingClient(timeout=20)
        cls._cls = ("obspy.clients.fdsn.routing.eidaws_routing_client."
                    "EIDAWSRoutingClient")

    def test_response_splitting(self):
        data = """
http://geofon.gfz.de/fdsnws/station/1/query
NU * * * 2017-01-01T00:00:00 2017-01-01T00:10:00

http://www.orfeus-eu.org/fdsnws/station/1/query
NS * * * 2017-01-01T00:00:00 2017-01-01T00:10:00
NR * * * 2017-01-01T00:00:00 2017-01-01T00:10:00
NO * * * 2017-01-01T00:00:00 2017-01-01T00:10:00
NL * * * 2017-01-01T00:00:00 2017-01-01T00:10:00
NA * * * 2017-01-01T00:00:00 2017-01-01T00:10:00

http://webservices.ingv.it/fdsnws/station/1/query
NI * * * 2017-01-01T00:00:00 2017-01-01T00:10:00

https://ws.resif.fr/fdsnws/station/1/query
ND * * * 2017-01-01T00:00:00 2017-01-01T00:10:00
        """.strip()
        # This should return a dictionary that contains the root URL of each
        # fdsn implementation and the POST payload ready to be submitted.
        assert EIDAWSRoutingClient._split_routing_response(data) == \
            {
                "http://geofon.gfz.de": (
                    "NU * * * 2017-01-01T00:00:00 2017-01-01T00:10:00"),
                "http://www.orfeus-eu.org": (
                    "NS * * * 2017-01-01T00:00:00 2017-01-01T00:10:00\n"
                    "NR * * * 2017-01-01T00:00:00 2017-01-01T00:10:00\n"
                    "NO * * * 2017-01-01T00:00:00 2017-01-01T00:10:00\n"
                    "NL * * * 2017-01-01T00:00:00 2017-01-01T00:10:00\n"
                    "NA * * * 2017-01-01T00:00:00 2017-01-01T00:10:00"),
                "http://webservices.ingv.it": (
                    "NI * * * 2017-01-01T00:00:00 2017-01-01T00:10:00"),
                "https://ws.resif.fr": (
                    "ND * * * 2017-01-01T00:00:00 2017-01-01T00:10:00")}

        data = """
http://geofon.gfz.de/fdsnws/station/1/query
NU * * * 2017-01-01T00:00:00 2017-01-01T00:10:00
        """.strip()
        # This should return a dictionary that contains the root URL of each
        # fdsn implementation and the POST payload ready to be submitted.
        assert EIDAWSRoutingClient._split_routing_response(data) == \
            {
                "http://geofon.gfz.de": (
                    "NU * * * 2017-01-01T00:00:00 2017-01-01T00:10:00")}

    def test_response_splitting_fdsnws_subdomain(self):
        data = """
http://eida.gein.noa.gr/fdsnws/station/1/
HP LTHK * * 2017-10-20T00:00:00 2599-12-31T23:59:59

https://data.raspberryshake.org/fdsnws/station/1/
AM RA14E * * 2017-10-20T00:00:00 2599-12-31T23:59:59
        """
        assert EIDAWSRoutingClient._split_routing_response(data) == \
            {"http://eida.gein.noa.gr":
                "HP LTHK * * 2017-10-20T00:00:00 2599-12-31T23:59:59",
             "https://data.raspberryshake.org":
                "AM RA14E * * 2017-10-20T00:00:00 2599-12-31T23:59:59"}

    def test_non_allowed_parameters(self):
        msg = 'The `filename` argument is not supported'
        with pytest.raises(ValueError, match=msg):
            self.client.get_waveforms(
                network="BW", station="ALTM", location="", channel="LHZ",
                starttime=obspy.UTCDateTime(2017, 1, 1),
                endtime=obspy.UTCDateTime(2017, 1, 1, 0, 1),
                filename="out.mseed")

        msg = 'The `attach_response` argument is not supported'
        with pytest.raises(ValueError, match=msg):
            self.client.get_waveforms(
                network="BW", station="ALTM", location="", channel="LHZ",
                starttime=obspy.UTCDateTime(2017, 1, 1),
                endtime=obspy.UTCDateTime(2017, 1, 1, 0, 1),
                attach_response=True)

        msg = 'The `filename` argument is not supported'
        with pytest.raises(ValueError, match=msg):
            self.client.get_waveforms_bulk([], filename="out.mseed")

        msg = 'The `attach_response` argument is not supported'
        with pytest.raises(ValueError, match=msg):
            self.client.get_waveforms_bulk([], attach_response=True)

        msg = 'The `filename` argument is not supported'
        with pytest.raises(ValueError, match=msg):
            self.client.get_stations(filename="out.xml")

        msg = 'The `filename` argument is not supported'
        with pytest.raises(ValueError, match=msg):
            self.client.get_stations_bulk([], filename="out.xml")

    def test_get_waveforms(self):
        """
        This just dispatches to the get_waveforms_bulk() method - so no need
        to also test it explicitly.
        """
        with mock.patch(self._cls + ".get_waveforms_bulk") as p:
            p.return_value = "1234"
            st = self.client.get_waveforms(
                network="XX", station="XXXXX", location="XX", channel="XXX",
                starttime=obspy.UTCDateTime(2017, 1, 1),
                endtime=obspy.UTCDateTime(2017, 1, 2),
                longestonly=True, minimumlength=2)
        assert st == "1234"
        assert p.call_count == 1
        assert p.call_args[0][0][0] == \
            ["XX", "XXXXX", "XX", "XXX", obspy.UTCDateTime(2017, 1, 1),
             obspy.UTCDateTime(2017, 1, 2)]
        assert p.call_args[1] == {"longestonly": True, "minimumlength": 2}

    def test_get_waveforms_bulk(self):
        # Some mock routing response.
        content = """
http://example1.com/fdsnws/station/1/query
AA B1 -- DD 2017-01-01T00:00:00 2017-01-02T00:10:00

http://example2.com/fdsnws/station/1/query
AA B2 -- DD 2017-01-01T00:00:00 2017-01-02T00:10:00
"""
        if hasattr(content, "encode"):
            content = content.encode()

        with mock.patch(self._cls + "._download") as p1, \
                mock.patch(self._cls + "._download_waveforms") as p2, \
                mock.patch(self._cls + ".get_stations_bulk") as p3:
            p1.return_value = _DummyResponse(content=content)
            p2.return_value = "1234"

            # For the underlying get_stations_bulk() call.
            _dummy_inv = mock.MagicMock()
            _dummy_inv.get_contents.return_value = {
                "channels": ["AA.BB.CC.DD", "AA.BB.CC.DD"]}
            p3.return_value = _dummy_inv

            st = self.client.get_waveforms_bulk(
                [["AA", "B*", "", "DD", obspy.UTCDateTime(2017, 1, 1),
                  obspy.UTCDateTime(2017, 1, 2)]],
                longestonly=True, minimumlength=2)
        assert st == "1234"

        assert p1.call_count == 1
        assert p1.call_args[0][0] == \
            "http://www.orfeus-eu.org/eidaws/routing/1/query"
        # This has been modified by our mocked call to get_stations_bulk().
        assert p1.call_args[1]["data"] == (
            b"service=dataselect\nformat=post\n"
            b"AA BB CC DD 2017-01-01T00:00:00.000000 "
            b"2017-01-02T00:00:00.000000")

        # This is the final call to _download_waveforms() which is again
        # dependent on the dummy response to the _download() function.
        assert p2.call_args[0][0] == {
            "http://example1.com":
            "AA B1 -- DD 2017-01-01T00:00:00 2017-01-02T00:10:00",
            "http://example2.com":
            "AA B2 -- DD 2017-01-01T00:00:00 2017-01-02T00:10:00"}
        assert p2.call_args[1] == {"longestonly": True, "minimumlength": 2}

        # Call to this only dependent on the original bulk request.
        assert p3.call_count == 1
        assert p3.call_args[0][0][0] == \
            ["AA", "B*", "--", "DD",
             str(obspy.UTCDateTime(2017, 1, 1))[:-1],
             str(obspy.UTCDateTime(2017, 1, 2))[:-1]]
        # Everything should be passed on.
        assert p3.call_args[1] == {
            "level": "channel", "longestonly": True, "minimumlength": 2}

    def test_get_stations(self):
        # Some mock routing response.
        content = """
http://example1.com/fdsnws/station/1/query
AA B1 -- DD 2017-01-01T00:00:00 2017-01-02T00:10:00

http://example2.com/fdsnws/station/1/query
AA B2 -- DD 2017-01-01T00:00:00 2017-01-02T00:10:00
"""
        if hasattr(content, "encode"):
            content = content.encode()

        with mock.patch(self._cls + "._download") as p1, \
                mock.patch(self._cls + "._download_stations") as p2:
            p1.return_value = _DummyResponse(content=content)
            p2.return_value = "1234"

            inv = self.client.get_stations(
                network="AA", channel="DD",
                starttime=obspy.UTCDateTime(2017, 1, 1), latitude=0.0,
                longitude=1.0)
        assert inv == "1234"

        assert p1.call_count == 1
        assert p1.call_args[0][0] == \
            "http://www.orfeus-eu.org/eidaws/routing/1/query"
        # Only a few arguments should be part of the URL.
        assert p1.call_args[1] == {
            'content_type': 'text/plain',
            'data': b'service=station\nformat=post\nalternative=false\n'
                    b'AA * * DD 2017-01-01T00:00:00.000000 *'}

        assert p2.call_args[0][0] == {
            "http://example1.com":
                "AA B1 -- DD 2017-01-01T00:00:00 2017-01-02T00:10:00",
            "http://example2.com":
                "AA B2 -- DD 2017-01-01T00:00:00 2017-01-02T00:10:00"}
        # SNCL and times should be filtered out.
        assert p2.call_args[1] == {
            "longitude": 1.0, "latitude": 0.0}

    def test_get_stations_bulk(self):
        # Some mock routing response.
        content = """
http://example1.com/fdsnws/station/1/query
AA B1 -- DD 2017-01-01T00:00:00 2017-01-02T00:10:00

http://example2.com/fdsnws/station/1/query
AA B2 -- DD 2017-01-01T00:00:00 2017-01-02T00:10:00
"""
        if hasattr(content, "encode"):
            content = content.encode()

        with mock.patch(self._cls + "._download") as p1, \
                mock.patch(self._cls + "._download_stations") as p2:
            p1.return_value = _DummyResponse(content=content)
            p2.return_value = "1234"

            inv = self.client.get_stations_bulk(
                [["AA", "B*", "", "DD", obspy.UTCDateTime(2017, 1, 1),
                  obspy.UTCDateTime(2017, 1, 2)]],
                latitude=0.0, longitude=1.0,
                starttime=obspy.UTCDateTime(2017, 1, 1))
        assert inv == "1234"

        assert p1.call_count == 1
        assert p1.call_args[0][0] == \
            "http://www.orfeus-eu.org/eidaws/routing/1/query"
        assert p1.call_args[1]["data"] == (
            b"service=station\nformat=post\nalternative=false\n"
            b"AA B* -- DD 2017-01-01T00:00:00.000000 "
            b"2017-01-02T00:00:00.000000")

        assert p2.call_args[0][0] == {
            "http://example1.com":
                "AA B1 -- DD 2017-01-01T00:00:00 2017-01-02T00:10:00",
            "http://example2.com":
                "AA B2 -- DD 2017-01-01T00:00:00 2017-01-02T00:10:00"}
        # Only select parameters are passed on.
        assert p2.call_args[1] == {
            "longitude": 1.0, "latitude": 0.0,
            "starttime": obspy.UTCDateTime(2017, 1, 1)}



# -*- coding: utf-8 -*-
"""
obspy.io.pdas - PDAS file read support for ObsPy
================================================

The obspy.io.pdas package contains methods in order to read files in the PDAS
file format.

:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)

"""
from .core import _is_pdas, _read_pdas


if __name__ == '__main__':
    import doctest
    doctest.testmod(exclude_empty=True)



# -*- coding: utf-8 -*-
import numpy as np

from obspy import Stream, UTCDateTime
from obspy.io.pdas import _is_pdas, _read_pdas


class TestPDAS():
    """
    Test suite for pdas.
    """
    def test_read_pdas(self, testdata):
        """
        Tests the _read_pdas function.
        """
        st = _read_pdas(testdata['p1246001.108'])
        assert isinstance(st, Stream)
        assert len(st) == 1
        tr = st[0]
        expected = [('COMMENT', 'GAINRANGED'),
                    ('DATASET', 'P1246001108'),
                    ('FILE_TYPE', 'LONG'),
                    ('HORZ_UNITS', 'Sec'),
                    ('SIGNAL', 'Channel1'),
                    ('VERSION', 'next'),
                    ('VERT_UNITS', 'Counts')]
        assert sorted(tr.stats.pop("pdas").items()) == expected
        expected = [('_format', 'PDAS'),
                    (u'calib', 1.0),
                    (u'channel', u''),
                    (u'delta', 0.005),
                    (u'endtime', UTCDateTime(1994, 4, 18, 0, 0, 2, 495000)),
                    (u'location', u''),
                    (u'network', u''),
                    (u'npts', 500),
                    (u'sampling_rate', 200.0),
                    (u'starttime', UTCDateTime(1994, 4, 18, 0, 0)),
                    (u'station', u'')]
        assert sorted(tr.stats.items()) == expected
        expected = np.array([895, 867, 747, 591, 359, -129, -185, 3, 115, 243],
                            dtype=np.int16)
        np.testing.assert_array_equal(tr.data[:10], expected)

    def test_is_pdas(self, testdata):
        """
        Tests the _read_pdas function.
        """
        assert _is_pdas(testdata['p1246001.108'])



# -*- coding: utf-8 -*-
from .blockette import Blockette
from ..fields import FixedString, Integer, VariableString


class Blockette031(Blockette):
    """
    Blockette 031: Comment Description Blockette.

    Station operators, data collection centers, and data management centers
    can add descriptive comments to data to indicate problems encountered or
    special situations.

    Sample:
    03100720750Stime correction does not include leap second, (-1000ms).~000
    """
    id = 31
    name = "Comment Description"
    fields = [
        Integer(3, "Comment code key", 4),
        FixedString(4, "Comment class code", 1),
        VariableString(5, "Description of comment", 1, 70, 'UNLPS'),
        Integer(6, "Units of comment level", 3, ignore=True)
    ]



# -*- coding: utf-8 -*-
"""
USAGE: make_assets.py [-f] [-c]
"""
import os
import shutil
import sys
from urllib import request


CDN_URL = 'https://netdna.bootstrapcdn.com/bootstrap/3.1.1/'

ASSETS = {
    'source/_static/css/base.css': 'http://tests.obspy.org/static/base.css',

    'source/_static/font.css':
        'http://tests.obspy.org/static/font/style.css',
    'source/_static/fonts/icomoon.eot':
        'http://tests.obspy.org/static/font/fonts/icomoon.eot',
    'source/_static/fonts/icomoon.svg':
        'http://tests.obspy.org/static/font/fonts/icomoon.svg',
    'source/_static/fonts/icomoon.ttf':
        'http://tests.obspy.org/static/font/fonts/icomoon.ttf',
    'source/_static/fonts/icomoon.woff':
        'http://tests.obspy.org/static/font/fonts/icomoon.woff',

    'source/_templates/navbar-local.html':
        'http://tests.obspy.org/snippets/navbar.html',
    'source/_templates/footer.html':
        'http://tests.obspy.org/snippets/footer.html',

    'source/_static/css/bootstrap.min.css':
        CDN_URL + 'css/bootstrap.min.css',
    'source/_static/fonts/glyphicons-halflings-regular.eot':
        CDN_URL + 'fonts/glyphicons-halflings-regular.eot',
    'source/_static/fonts/glyphicons-halflings-regular.svg':
        CDN_URL + 'fonts/glyphicons-halflings-regular.svg',
    'source/_static/fonts/glyphicons-halflings-regular.ttf':
        CDN_URL + 'fonts/glyphicons-halflings-regular.ttf',
    'source/_static/fonts/glyphicons-halflings-regular.woff':
        CDN_URL + 'fonts/glyphicons-halflings-regular.woff',
}

force = '-f' in sys.argv
clean = '-c' in sys.argv

if clean:
    print('Cleaning assets ...')
elif force:
    print('Forced downloading assets ...')
else:
    print('Downloading necessary assets ...')

for asset, url in ASSETS.items():
    if clean:
        try:
            print('Deleting %s ...' % (asset))
            os.remove(asset)
        except Exception:
            if force:
                pass
            else:
                raise

    elif force or not os.path.exists(asset):
        print('Downloading %s ...' % (url))
        resp = request.urlopen(url)
        with open(asset, 'wb') as output:
            shutil.copyfileobj(resp, output)
        resp.close()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
The obspy.io.win.core test suite.
"""
from obspy import read
from obspy.core.utcdatetime import UTCDateTime
from obspy.io.win.core import _read_win


class TestCore():
    """
    Test cases for win core interface
    """
    def test_read_via_obspy(self, testdata):
        """
        Read files via obspy.core.stream.read function.
        """
        filename = testdata['10030302.00']
        # 1
        st = read(filename)
        st.verify()
        st.sort(keys=['channel'])
        assert len(st) == 2
        assert st[0].stats.starttime == \
            UTCDateTime('2010-03-03T02:00:00.000000Z')
        assert st[0].stats.endtime == \
            UTCDateTime('2010-03-03T02:00:59.990000Z')
        assert st[0].stats.starttime == \
            UTCDateTime('2010-03-03T02:00:00.000000Z')
        assert len(st[0]) == 6000
        assert round(abs(st[0].stats.sampling_rate-100.0), 7) == 0
        assert st[0].stats.channel == 'a100'

    def test_read_via_module(self, testdata):
        """
        Read files via obspy.io.win.core._read_win function.
        """
        filename = testdata['10030302.00']
        # 1
        st = _read_win(filename)
        st.verify()
        st.sort(keys=['channel'])
        assert len(st) == 2
        assert st[0].stats.starttime == \
            UTCDateTime('2010-03-03T02:00:00.000000Z')
        assert st[0].stats.endtime == \
            UTCDateTime('2010-03-03T02:00:59.990000Z')
        assert st[0].stats.starttime == \
            UTCDateTime('2010-03-03T02:00:00.000000Z')
        assert len(st[0]) == 6000
        assert round(abs(st[0].stats.sampling_rate-100.0), 7) == 0
        assert st[0].stats.channel == 'a100'

    def test_read_05_byte_data(self, testdata):
        """
        Read a file included 0.5 byte datawide
        """
        filename = testdata['1070533011_1701260003.win']

        st = _read_win(filename)
        st.sort(keys=['channel'])
        assert len(st[0]) == 6000
        assert len(st[1]) == 6000
        assert len(st[2]) == 6000
        assert st[0].stats.starttime == \
            UTCDateTime('2017-01-26T00:03:00.000000Z')
        assert st[0].stats.endtime == \
            UTCDateTime('2017-01-26T00:03:59.990000Z')
        assert st[1].stats.starttime == \
            UTCDateTime('2017-01-26T00:03:00.000000Z')
        assert st[1].stats.endtime == \
            UTCDateTime('2017-01-26T00:03:59.990000Z')
        assert st[2].stats.starttime == \
            UTCDateTime('2017-01-26T00:03:00.000000Z')
        assert st[2].stats.endtime == \
            UTCDateTime('2017-01-26T00:03:59.990000Z')

    def test_read_1kHz_data(self, testdata):
        """ Reads a file with 1kHz sampling rate
        Was an issue with header parsing, see #3641"""
        filename = testdata['25112616_ch0000.10']
        st = _read_win(filename)
        assert st[0].stats.sampling_rate == 1000.0

    def test_read_24bit_data(self, testdata):
        """ Reads a file with 24bit data """
        filename = testdata['25112618_ch0000.24bits']
        st = read(filename)
        assert st[0].stats.sampling_rate == 200.0

        expected_stds = {
            "...0000": 62491.2874698}

        for tr in st:
            assert round(expected_stds[tr.id], 4) == round(tr.data.std(), 4)



# -*- coding: utf-8 -*-
"""
Decorator used in ObsPy.

:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
import functools
import inspect
from pathlib import Path
import re
import socket
import tarfile
import warnings
import zipfile

import numpy as np
from decorator import decorator

from obspy.core.util import get_example_file
from obspy.core.util.base import NamedTemporaryFile
from obspy.core.util.deprecation_helpers import ObsPyDeprecationWarning


def deprecated(warning_msg=None):
    """
    This is a decorator which can be used to mark functions as deprecated.

    .. note::
        Actually, this is not a decorator itself but a decorator factory,
        returning the correct decorator for the specified options. It can be
        used just like a decorator.

    It will result in a warning being emitted when the function is used.
    """
    @decorator
    def _deprecated(func, *args, **kwargs):
        if 'deprecated' in str(func.__doc__).lower():
            msg = func.__doc__
        elif warning_msg:
            msg = warning_msg
            func.__doc__ = warning_msg
        else:
            msg = "Call to deprecated function %s." % func.__name__
        warnings.warn(msg, category=ObsPyDeprecationWarning, stacklevel=3)
        return func(*args, **kwargs)
    return _deprecated


def deprecated_keywords(keywords):
    """
    Decorator for marking keywords as deprecated.

    .. note::
        Actually, this is not a decorator itself but a decorator factory,
        returning the correct decorator for the specified options. It can be
        used just like a decorator.

    :type keywords: dict
    :param keywords: old/new keyword names as key/value pairs.
    """
    def fdec(func):
        fname = func.__name__
        msg = "Deprecated keyword %s in %s() call - please use %s instead."
        msg2 = "Deprecated keyword %s in %s() call - ignoring."
        msg3 = ("Conflicting deprecated keywords (%s) in %s() call"
                " - please use new '%s' keyword instead.")

        @functools.wraps(func)
        def echo_func(*args, **kwargs):
            # check if multiple deprecated keywords get mapped to the same new
            # keyword
            new_keyword_appearance_counts = dict.fromkeys(keywords.values(), 0)
            for key, new_key in keywords.items():
                if key in kwargs:
                    new_keyword_appearance_counts[new_key] += 1
            for key_ in keywords.values():
                # ignore `None` as new value, it means that no mapping is
                # happening..
                if key_ is None:
                    continue
                if new_keyword_appearance_counts[key_] > 1:
                    conflicting_keys = ", ".join(
                        [old_key for old_key, new_key in keywords.items()
                         if new_key == key_])
                    raise Exception(msg3 % (conflicting_keys, fname, new_key))
            # map deprecated keywords to new keywords
            for kw in list(kwargs):
                if kw in keywords:
                    nkw = keywords[kw]
                    if nkw is None:
                        warnings.warn(msg2 % (kw, fname),
                                      category=ObsPyDeprecationWarning,
                                      stacklevel=3)
                    else:
                        warnings.warn(msg % (kw, fname, nkw),
                                      category=ObsPyDeprecationWarning,
                                      stacklevel=3)
                        kwargs[nkw] = kwargs[kw]
                    del kwargs[kw]
            return func(*args, **kwargs)
        return echo_func

    return fdec


@decorator
def skip_on_network_error(func, *args, **kwargs):
    """
    Decorator to mark test routines that fail with certain network
    errors (e.g. timeouts) as "skipped" rather than "Error".
    """
    import pytest  # NOQA

    try:
        return func(*args, **kwargs)
    ###################################################
    # add more except clauses like this to add other
    # network errors that should be skipped
    except socket.timeout as e:
        if str(e) == "timed out":
            pytest.skip(str(e))
    ###################################################
    except socket.error as e:
        if str(e) == "[Errno 110] Connection timed out":
            pytest.skip(str(e))
    # general except to be able to generally reraise
    except Exception:
        raise


@decorator
def uncompress_file(func, filename, *args, **kwargs):
    """
    Decorator used for temporary uncompressing file if .gz or .bz2 archive.
    """
    if not kwargs.pop('check_compression', True):
        return func(filename, *args, **kwargs)
    if not isinstance(filename, str):
        return func(filename, *args, **kwargs)
    elif not Path(filename).exists():
        msg = "File not found '%s'" % (filename)
        raise IOError(msg)
    # check if we got a compressed file or archive
    obj_list = []
    if tarfile.is_tarfile(filename):
        try:
            # reading with transparent compression
            with tarfile.open(filename, 'r|*') as tar:
                for tarinfo in tar:
                    # only handle regular files
                    if not tarinfo.isfile():
                        continue
                    data = tar.extractfile(tarinfo).read()
                    # Skip empty files - we don't need them no matter what
                    # and it guards against rare cases where waveforms files
                    # are also slightly valid tar-files.
                    if not data:
                        continue
                    obj_list.append(data)
        except Exception:
            pass
    elif zipfile.is_zipfile(filename):
        try:
            with zipfile.ZipFile(filename) as zip:
                if b'obspy_no_uncompress' in zip.comment:
                    # be nice to plugins based on zip format
                    # do not uncompress the file if tag is present
                    # see issue #3192
                    obj_list = None
                else:
                    obj_list = [zip.read(name) for name in zip.namelist()]
        except Exception:
            pass
    elif filename.endswith('.bz2'):
        # bz2 module
        try:
            import bz2
            with open(filename, 'rb') as fp:
                obj_list.append(bz2.decompress(fp.read()))
        except Exception:
            pass
    elif filename.endswith('.gz'):
        # gzip module
        try:
            import gzip
            with gzip.open(filename, 'rb') as fp:
                obj_list.append(fp.read())
        except Exception:
            pass
    # handle results
    if obj_list:
        # write results to temporary files
        result = None
        for obj in obj_list:
            with NamedTemporaryFile() as tempfile:
                tempfile._fileobj.write(obj)
                stream = func(tempfile.name, *args, **kwargs)
                # just add other stream objects to first stream
                if result is None:
                    result = stream
                else:
                    result += stream
    else:
        # no compressions
        result = func(filename, *args, **kwargs)
    return result


@decorator
def raise_if_masked(func, *args, **kwargs):
    """
    Raises if the first argument (self in case of methods) is a Trace with
    masked values or a Stream containing a Trace with masked values.
    """
    arrays = []
    # first arg seems to be a Stream
    if hasattr(args[0], "traces"):
        arrays = [tr.data for tr in args[0]]
    # first arg seems to be a Trace
    if hasattr(args[0], "data") and isinstance(args[0].data, np.ndarray):
        arrays = [args[0].data]
    for arr in arrays:
        if np.ma.is_masked(arr):
            msg = "Trace with masked values found. This is not " + \
                  "supported for this operation. Try the split() " + \
                  "method on Trace/Stream to produce a Stream with " + \
                  "unmasked Traces."
            raise NotImplementedError(msg)
    return func(*args, **kwargs)


@decorator
def skip_if_no_data(func, *args, **kwargs):
    """
    Does nothing if the first argument (self in case of methods) is a Trace
    with no data in it.
    """
    if not args[0]:
        return
    return func(*args, **kwargs)


def map_example_filename(arg_kwarg_name):
    """
    Decorator that replaces "/path/to/filename" patterns in the arg or kwarg
    of the specified name with the correct file path. If the pattern is not
    encountered nothing is done.

    .. note::
        Actually, this is not a decorator itself but a decorator factory,
        returning the correct decorator for the specified options. It can be
        used just like a decorator.

    :type arg_kwarg_name: str
    :param arg_kwarg_name: name of the arg/kwarg that should be (tried) to map
    """
    @decorator
    def _map_example_filename(func, *args, **kwargs):
        prefix = '/path/to/'
        # check kwargs
        if arg_kwarg_name in kwargs:
            if isinstance(kwargs[arg_kwarg_name], str):
                if re.match(prefix, kwargs[arg_kwarg_name]):
                    try:
                        kwargs[arg_kwarg_name] = \
                            get_example_file(kwargs[arg_kwarg_name][9:])
                    # file not found by get_example_file:
                    except IOError:
                        pass
        # check args
        else:
            try:
                inspected_args = [
                    p.name
                    for p in inspect.signature(func).parameters.values()
                ]
            except AttributeError:
                inspected_args = inspect.getargspec(func).args
            try:
                ind = inspected_args.index(arg_kwarg_name)
            except ValueError:
                pass
            else:
                if ind < len(args) and isinstance(args[ind], str):
                    # need to check length of args from inspect
                    if re.match(prefix, args[ind]):
                        try:
                            args = list(args)
                            args[ind] = get_example_file(args[ind][9:])
                            args = tuple(args)
                        # file not found by get_example_file:
                        except IOError:
                            pass
        return func(*args, **kwargs)
    return _map_example_filename


if __name__ == '__main__':
    import doctest
    doctest.testmod(exclude_empty=True)



# -*- coding: utf-8 -*-
"""
Evt (Kinemetrics) format support for ObsPy.

:copyright:
    Royal Observatory of Belgium, 2013
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
from struct import unpack
import warnings

import numpy as np

from obspy import Stream, Trace
from obspy.core.compatibility import from_buffer
from obspy.core.util.deprecation_helpers import ObsPyDeprecationWarning
from .evt_base import (EvtBadDataError, EvtBadHeaderError, EvtEOFError,
                       EvtVirtual)

WARNING_HEADER = "Only tested with files from ROB networks :" + \
    " - New Etna and old Etna" + \
    " - ByteOrder : MSB first (Motorola) " + \
    " - File Header of 2040 bytes (12 Channels)" + \
    " - .........." + \
    "Other situation may not work (not yet implemented)."

FRAME_STRUCT = b"BBHHLHHBBHB13s"

# 0 - 0x7c (0-34)
HEADER_STRUCT1 = b"3sB2H3B3x6Hh2Hh22x" + b"3B5x2H4hH2x2h6L16x"
# 0x7c - 0x22c (35-106)
HEADER_STRUCT2 = b"lLlL2l12x" * 12
# 0x22c-0x2c8(->139)
HEADER_STRUCT3 = b"3L4H2L4x2H5s33sh2f4h4B2LB17s2B2B6xh22x"
# 0x2c8 - 0x658 (140 - 464) (27*12)
HEADER_STRUCT4 = b"5sbH5h3H4BHBx8f2B10x" * 12
# 0x658 - 0x688 items 167-184
HEADER_STRUCT5 = b"3B5x6H2hx2Bxh3Hl8x"
# 0x688 - 0x6c4 (3*15)
HEADER_STRUCT6 = b"cBh" * 15
# 0x6c4 - 0x7f8
HEADER_STRUCT7 = b"64s16s16s16s16s16s24s24s24s24s3B3b5h4xH46s"

structures = {}
structures["instrument"] = {0: 'QDR', 9: 'K2', 10: 'Makalu', 20: 'New Etna',
                            30: 'Rock', 40: 'SSA2EVT'}
structures["stream_flags"] = {0: "functional test, FT",
                              1: "sensor response test, SRT, 080129",
                              2: "recorded trigger data, TRIG DATA"}
structures["restart_source"] = {0: "unknown",
                                1: "power switch",
                                2: "user command",
                                3: "software watchdog",
                                4: "DSP failure",
                                5: "battery failure",
                                6: "memory error"}
structures["header_version"] = {100: "12 channel, old",
                                110: "12 channel, old",
                                120: "18 channel, old",
                                130: "12 channel, old",
                                140: "12 channel, K2, Etna, QDR, Rock",
                                150: "18 channel, Mt Whitney, Rock"}
structures["clocksource"] = {0: "RTC from cold start",
                             1: "keyboard",
                             2: "Sync w/ ext. ref. pulse",
                             3: "Internal GPS"}
structures["chan_triggertype"] = {0: "threshold, default",
                                  1: "sta/lta"}
structures["chan_iirtrigfilter"] = {0: "iira IIR bandpass, 1.2 to 20Hz "
                                       "@ 200sps",
                                    1: "CSM, classic strong motion, 0.1 to "
                                       "12.5 Hz @ 200sps",
                                    2: "iirc IIR bandpass, 2.0 to 40Hz at "
                                       "200sps"}
structures["chan_stasecondstten"] = {0: 0.1,
                                     1: 0.2,
                                     2: 0.3,
                                     3: 0.4,
                                     4: 0.5,
                                     5: 0.6,
                                     6: 0.8,
                                     7: 1.0,
                                     8: 1.2,
                                     9: 1.4,
                                     'a': 1.6,
                                     'b': 1.8,
                                     'c': 2.0,
                                     'd': 2.5,
                                     'e': 3.0,
                                     'f': 5.0,
                                     10: 10.0}
structures["chan_ltaseconds"] = {0: 20,
                                 1: 30,
                                 2: 40,
                                 3: 50,
                                 4: 60,
                                 5: 80,
                                 6: 100,
                                 7: 120}
structures["chan_staltaratio"] = {0: 1.5,
                                  1: 2,
                                  2: 3,
                                  3: 4,
                                  4: 6,
                                  5: 8,
                                  6: 10,
                                  7: 15,
                                  8: 20,
                                  9: 30,
                                  'a': 40,
                                  'b': 60,
                                  'c': 100}
structures["chan_staltaprecent"] = {0: 10,
                                    1: 15,
                                    2: 20,
                                    3: 40,
                                    4: 60,
                                    5: 100}
structures["chan_range"] = {0: 'not defined',
                            1: '4g',
                            2: '2g',
                            3: '1g',
                            4: '1/2g',
                            5: '1/4g',
                            6: '1/8g',
                            7: '1/16g',
                            8: '1/32g',
                            9: '1/64g',
                            'a': '1/128g',
                            'b': '1/256g',
                            'c': '1/512g',
                            'd': '1/1024g',
                            }
structures["voter_type"] = {'C': 'channel',
                            'E': 'external',
                            'K': 'keyboard',
                            'N': 'network',
                            '': ''}


class Evt(object):
    """
    Class to read Evt (Kinemetrics) formatted files.
    """
    def __init__(self):
        self.e_tag = EvtTag()
        self.e_header = EvtHeader()
        self.e_frame = EvtFrameHeader()
        self.e_data = EvtData()
        self.samplingrate = 0

    def get_calibs(self):
        """
        Gets calibration values for each traces

        Note about calibration:
            fullscale of instrument = +/- 2.5V & +/-20V
            data : 4 bytes - 24 bits
            calibration in volts = data * fullscale / 2**23

            sensitivity = volts/g
            calibration in MKS units = (data_in_volts / sensitivity) * g

        """
        calibs = []
        for i in range(self.e_header.nchannels):
            calib_volts = 8388608.0 / self.e_header.chan_fullscale[i]
            # 8388608 = 2**23
            calib_mks = (calib_volts *
                         self.e_header.chan_sensitivity[i]) / (9.81)
            # 9.81 = mean value of g on earth
            calibs.append(calib_mks)
        return calibs

    def calibration(self):
        """
        Apply calibrations on data matrix
        """
        calibs = self.get_calibs()
        for i in range(self.e_header.nchannels):
            self.data[i] /= calibs[i]

    def read_file(self, filename_or_object, apply_calib=False,
                  encoding=None, **kwargs):
        """
        Reads an Evt file to the internal data structure

        :type filename_or_object: str or file-like object
        :param filename_or_object: Evt file to be read
        :type apply_calib: bool
        :param apply_calib: If False, the raw data will be returned (no
            corrections, int32, default), if True, the data will be in m/s2.
        :rtype: obspy.core.stream.Stream
        :return: Obspy Stream with data
        """

        # Need to deprecate "raw" until version 2.1
        if "raw" in kwargs:
            apply_calib = not kwargs["raw"]
            warnings.warn(
                'The "raw" keyword argument is deprecated, please use '
                '"apply_calib" instead. Note the behaviour is flipped, so '
                '"raw=True" becomes "apply_calib=False, which is default now. '
                'Setting "apply_calib=%s" and continuing...' % apply_calib,
                ObsPyDeprecationWarning)

        if encoding:
            self.e_header.ENCODING = encoding

        # Support reading from filenames of file-like objects.
        if hasattr(filename_or_object, "seek") and \
                hasattr(filename_or_object, "tell") and \
                hasattr(filename_or_object, "read"):
            is_fileobject = True
            file_pointer = filename_or_object
        else:
            is_fileobject = False
            file_pointer = open(filename_or_object, "rb")

        try:
            self.e_tag.read(file_pointer)
            endian = self.e_tag.endian
            self.e_header.unset_dict()
            self.e_header.read(file_pointer, self.e_tag.length, endian)

            self.data = np.ndarray([self.e_header.nchannels, 0])

            while True:
                try:
                    self.e_tag.read(file_pointer)
                    retparam = self.e_frame.read(file_pointer,
                                                 self.e_tag.length, endian)
                    if self.samplingrate == 0:
                        self.samplingrate = retparam[0]
                    elif self.samplingrate != retparam[0]:
                        raise EvtBadHeaderError("Sampling rate not constant")
                    data = self.e_data.read(file_pointer,
                                            self.e_tag.datalength,
                                            endian, retparam)
                    self.data = np.hstack((self.data, data))
                except EvtEOFError:
                    break
        finally:
            if not is_fileobject:
                file_pointer.close()

        if self.e_frame.count() != self.e_header.duration:
            raise EvtBadDataError("Bad number of blocks")

        calibs = self.get_calibs()
        if apply_calib:
            self.calibration()

        traces = []
        for i in range(self.e_header.nchannels):
            cur_trace = Trace(data=self.data[i])
            cur_trace.stats.channel = str(i)
            cur_trace.stats.station = self.e_header.stnid
            cur_trace.stats.sampling_rate = float(self.samplingrate)
            cur_trace.stats.starttime = self.e_header.starttime
            cur_trace.stats.kinemetrics_evt = self.e_header.make_obspy_dict(i)
            if not apply_calib:
                cur_trace.stats.calib = 1./calibs[i]
            else:
                cur_trace.stats.calib = 1.
            traces.append(cur_trace)

        return Stream(traces=traces)


class EvtData(object):
    """
    Class to manage data in Evt file
    """
    def read(self, file_p, length, endian, param):
        """
        read data from file_p

        :param file_p: file pointer
        :param length: length to be read
        :param endian: endian type in datafile
        :type param: list
        :param param: sampling rate,sample size, block time, channels
        :rtype: list of list
        :return: list of data
        """
        buff = file_p.read(length)
        samplerate = param[0]
        numbyte = param[1]
        numchan = param[3]
        num = (samplerate // 10) * numbyte * numchan
        if length != num:
            raise EvtBadDataError("Bad data length")

        if numbyte == 2:
            data = from_buffer(buff, ">h").reshape((-1, numchan)).T
        elif numbyte == 4:
            data = from_buffer(buff, ">i").reshape((-1, numchan)).T
        elif numbyte == 3:
            data = np.empty((numchan, samplerate // 10))
            for j in range(samplerate // 10):
                for k in range(numchan):
                    i = (j * numchan) + k
                    val = unpack(b">i", buff[i * 3:(i * 3) + 3] + b'\0')[0] \
                        >> 8
                    data[k, j] = val

        return data


class EvtHeader(EvtVirtual):
    """
    Class to manage header of Evt file
    """
    ENCODING = "latin-1"
    HEADER = {'id': [0, ['_strnull', '']],
              'instrument': [1, ''],
              'headerversion': [2, ''],
              'headerbytes': [3, ''],
              'a2dbits': [4, ''],
              'samplebytes': [5, ''],
              'restartsource': [6, ''],
              'installedchan': [7, ''],
              'maxchannels': [8, ''],
              'sysblkversion': [9, ''],
              'bootblkversion': [10, ''],
              'appblkversion': [11, ''],
              'dspblkversion': [12, ''],
              'batteryvoltage': [13, ''],
              'crc': [14, ''],
              'flags': [15, ''],
              'temperature': [16, ''],
              'clocksource': [17, ''],
              'gpsstatus': [18, ['_gpsstatus', '']],
              'gpssoh': [19, ''],
              'gpslockfailcount': [20, ''],
              'gpsupdatertccount': [21, ''],
              'acqdelay': [22, ''],
              'gpslatitude': [23, ''],
              'gpslongitude': [24, ''],
              'gpsaltitude': [25, ''],
              'daccount': [26, ''],
              'gpslastdrift1': [27, ''],
              'gpslastdrift2': [28, ''],
              'gpslastturnontime1': [29, ['_time', 0]],
              'gpslastturnontime2': [30, ['_time', 0]],
              'gpslastupdatetime1': [31, ['_time', 0]],
              'gpslastupdatetime2': [32, ['_time', 0]],
              'gpslastlock1': [33, ['_time', 0]],
              'gpslastlock2': [34, ['_time', 0]],
              'maxpeak': [35, ['_array', [12, 6, 35]]],
              'maxpeakoffset': [36, ['_array', [12, 6, 36]]],
              'minpeak': [37, ['_array', [12, 6, 37]]],
              'minpeakoffset': [38, ['_array', [12, 6, 38]]],
              'mean': [39, ['_array', [12, 6, 39]]],
              'aqoffset': [40, ['_array', [12, 6, 40]]],
              'starttime': [107, ['_time', 112]],
              'triggertime': [108, ['_time', 113]],
              'duration': [109, ''],
              'errors': [110, ''],
              'stream_flags': [111, ''],
              'starttimemsec': [112, ''],
              'triggertimemsec': [113, ''],
              'nscans': [114, ''],
              'triggerbitmap': [115, ''],
              'serialnumber': [116, ''],
              'nchannels': [117, ''],
              'stnid': [118, ['_strnull', '']],
              'comment': [119, ['_strnull', '']],
              'elevation': [120, ''],
              'latitude': [121, ''],
              'longitude': [122, ''],
              'gpsturnoninterval': [137, ''],
              'gpsmaxturnontime': [138, ''],
              'localoffset': [139, ''],
              'chan_id': [140, ['_arraynull', [12, 27, 140]]],
              'chan_channel': [141, ['_array', [12, 27, 141]]],
              'chan_sensorserialnumberext': [141, ['_array', [12, 27, 142]]],
              'chan_north': [143, ['_array', [12, 27, 143]]],
              'chan_east': [144, ['_array', [12, 27, 144]]],
              'chan_up': [145, ['_array', [12, 27, 145]]],
              'chan_altitude': [146, ['_array', [12, 27, 146]]],
              'chan_azimuth': [147, ['_array', [12, 27, 147]]],
              'chan_sensortype': [148, ['_array', [12, 27, 148]]],
              'chan_sensorserialnumber': [149, ['_array', [12, 27, 149]]],
              'chan_gain': [150, ['_array', [12, 27, 150]]],
              'chan_triggertype': [151, ['_array', [12, 27, 151]]],
              'chan_iirtrigfilter': [152, ['_array', [12, 27, 152]]],
              'chan_stasecondstten': [153, ['_array', [12, 27, 153]]],
              'chan_ltaseconds': [154, ['_array', [12, 27, 154]]],
              'chan_staltaratio': [155, ['_array', [12, 27, 155]]],
              'chan_staltaprecent': [156, ['_array', [12, 27, 156]]],
              'chan_fullscale': [157, ['_array', [12, 27, 157]]],
              'chan_sensitivity': [158, ['_array', [12, 27, 158]]],
              'chan_damping': [159, ['_array', [12, 27, 159]]],
              'chan_natfreq': [160, ['_array', [12, 27, 160]]],
              'chan_calcoil': [164, ['_array', [12, 27, 164]]],
              'chan_range': [165, ['_array', [12, 27, 165]]],
              'chan_sensorgain': [166, ['_array', [12, 27, 166]]],

              'filterflag': [167, ''],  # header5, 167-184,
              'primarystorage': [168, ''],
              'secondarystorage': [169, ''],
              'eventnumber': [170, ''],
              'sps': [171, ''],
              'apw': [172, ''],
              'preevent': [173, ''],
              'postevent': [174, ''],
              'minruntime': [175, ''],
              'votestotrigger': [176, ''],
              'votestodetrigger': [177, ''],
              'filtertype': [178, ''],
              'datafmt': [179, ''],
              'timeout': [180, ''],
              'txblksize': [181, ''],
              'buffersize': [182, ''],
              'samplerate': [183, ''],
              'txchanmap': [184, ''],

              'voter_type': [185, ['_arraynull', [15, 3, 185]]],  # header6
              'voter_number': [186, ['_array', [15, 3, 186]]],
              'voter_weight': [187, ['_array', [15, 3, 187]]],

              'modem_initcmd': [188, ['_strnull', '']],  # header7
              'modem_dialingprefix': [189, ['_strnull', '']],
              'modem_dialingsuffix': [190, ['_strnull', '']],
              'modem_hangupcmd': [191, ['_strnull', '']],
              'modem_autoansweroncmd': [192, ['_strnull', '']],
              'modem_autoansweroffcmd': [193, ['_strnull', '']],
              'modem_phonenumber1': [194, ['_strnull', '']],
              'modem_phonenumber2': [195, ['_strnull', '']],
              'modem_phonenumber3': [196, ['_strnull', '']],
              'modem_phonenumber4': [197, ['_strnull', '']],
              'modem_waitforconnection': [198, ''],
              'modem_pausebetweencalls': [199, ''],
              'modem_maxdialattempts': [200, ''],
              'modem_cellshare': [201, ''],
              'modem_cellontime': [202, ''],
              'modem_cellwarmuptime': [203, ''],
              'modem_cellstarttime1': [204, ''],
              'modem_cellstarttime2': [205, ''],
              'modem_cellstarttime3': [206, ''],
              'modem_cellstarttime4': [207, ''],
              'modem_cellstarttime5': [208, ''],
              'modem_flags': [209, ''],
              'modem_calloutmsg': [210, ['_strnull', '']],
              }

    def __init__(self):
        EvtVirtual.__init__(self)

    def read(self, file_p, length, endian):
        """
        read the Header of Evt file
        """
        buff = file_p.read(length)
        self.endian = endian
        if length == 2040:  # File Header 12 channel
            self.analyse_header12(buff)
        elif length == 2736:  # File Header 18 channel
            raise NotImplementedError("16 Channel not implemented")
        else:
            raise EvtBadHeaderError("Bad Header length " + length)

    def analyse_header12(self, head_buff):
        val = unpack(self.endian + HEADER_STRUCT1, head_buff[0:0x7c])
        self.set_dict(val, 0, 34)
        val = unpack(self.endian + HEADER_STRUCT2, head_buff[0x7c:0x22c])
        self.set_dict(val, 35, 106)
        val = unpack(self.endian + HEADER_STRUCT3, head_buff[0x22c:0x2c8])
        self.set_dict(val, 107, 139)
        val = unpack(self.endian + HEADER_STRUCT4, head_buff[0x2c8:0x658])
        self.set_dict(val, 140, 166)
        val = unpack(self.endian+HEADER_STRUCT5, head_buff[0x658:0x688])
        self.set_dict(val, 167, 184)
        val = unpack(self.endian+HEADER_STRUCT6, head_buff[0x688:0x6c4])
        self.set_dict(val, 185, 187)
        val = unpack(self.endian+HEADER_STRUCT7, head_buff[0x6c4:0x7f8])
        self.set_dict(val, 188, 210)

        for key in self.HEADER:
            if key in structures:
                value = self.HEADER[key][2]
                if isinstance(value, list):
                    vals = []
                    for v in value:
                        try:
                            vals.append(structures[key][v])
                        except KeyError:
                            msg = f"Unmatched raw value: {v} (type={type(v)})"
                            warnings.warn(key+": "+msg)
                            vals.append(msg)
                    self.HEADER[key][2] = vals
                else:
                    try:
                        self.HEADER[key][2] = structures[key][value]
                    except KeyError:
                        msg = (f"Unmatched raw value: {value}"
                               f" (type={type(value)})")
                        warnings.warn(key + ": " + msg)
                        self.HEADER[key][2] = "Unmatched raw value: " \
                                              "{}".format(value)

    def make_obspy_dict(self, numchan):
        """
        Make an ObsPy dictionary from header dictionary for 1 channel

        :param numchan: channel to be converted
        :rtype: dictionary
        """
        dico = {}
        for key in self.HEADER:
            value = self.HEADER[key][2]
            if isinstance(value, list):
                dico[key] = value[numchan]
            else:
                dico[key] = value
        return dico


class EvtFrameHeader(EvtVirtual):
    """
    Class to manage frame header in Evt file
    """
    HEADER = {'frametype': [0, ''],
              'instrumentcode': [1, ''],
              'recorderid': [2, ''],
              'framesize': [3, ''],
              'blocktime': [4, ['_time', 9]],
              'channelbitmap': [5, ''],
              'streampar': [6, ''],
              'framestatus': [7, ''],
              'framestatus2': [8, ''],
              'msec': [9, ''],
              'channelbitmap1': [10, ''],
              'timecode': [11, '']}

    def __init__(self):
        EvtVirtual.__init__(self)
        self.numframe = 0
        self.endian = 0

    def count(self):
        """
        return the number of frames read
        """
        return self.numframe

    def read(self, file_p, length, endian):
        """
        read a frame

        :rtype: list
        :return: samplingrate, samplesize, blocktime, channels
        """
        buff = file_p.read(length)
        self.endian = endian
        if length == 32:  # Frame Header
            self.analyse_frame32(buff)
        else:
            raise EvtBadHeaderError("Bad Header length " + length)
        samplingrate = self.streampar & 4095
        samplesize = (self.framestatus >> 6) + 1
        if samplesize not in [2, 3, 4]:
            samplesize = 3  # default value = 3
        return (samplingrate, samplesize, self.blocktime, self.channels())

    def analyse_frame32(self, head_buff):
        self.numframe += 1
        val = unpack(self.endian + FRAME_STRUCT, head_buff)
        self.set_dict(val, 0)
        if not self.verify(verbose=False):
            raise EvtBadHeaderError("Bad Frame values")

    def verify(self, verbose=False):
        if self.frametype not in [3, 4, 5]:
            if verbose:
                print("FrameType ", self.frametype, " not Known")
            return False
        return True

    def channels(self):
        numchan = 12
        if self.frametype == 4:
            raise NotImplementedError("16 Channels not implemented")
        chan = 0
        for i in range(numchan):
            pow_of_2 = 2 ** i
            if self.channelbitmap & pow_of_2:
                chan += 1
        return chan


class EvtTag(EvtVirtual):
    """
    Class to read the TAGs of Evt Files
    """
    HEADER = {'order': [1, ''],
              'version': [2, ''],
              'instrument': [3, ''],
              'type': [4, ''],
              'length': [5, ''],
              'datalength': [6, ''],
              'id': [7, ''],
              'checksum': [8, ''],
              'endian': [9, '']}

    def __init__(self):
        EvtVirtual.__init__(self)
        self.endian = 0

    def read(self, file_p):
        """
        :type file_p: str
        :param file_p: file descriptor of Evt file.
        """
        mystr = file_p.read(16)
        if len(mystr) < 16:
            raise EvtEOFError
        sync, byte_order = unpack(b"cB", mystr[0:2])
        if sync == b'\x00':
            raise EvtEOFError
        if sync != b'K':
            raise EvtBadHeaderError('Sync error')
        if byte_order == 1:
            endian = b">"
        elif byte_order == 0:
            endian = b"<"
        else:
            raise EvtBadHeaderError
        val = unpack(endian + b"cBBBLHHHH", mystr)
        self.set_dict(val)
        self.endian = endian
        if not self.verify(verbose=False):
            raise EvtBadHeaderError("Bad Tag values")

    def verify(self, verbose=False):
        if self.type not in [1, 2]:
            if verbose:
                print("Type of Header ", self.type, " not known")
            return False
        if self.type == 1 and self.length not in [2040, 2736]:
            if verbose:
                print("Bad Header file length : ", self.length)
            return False
        return True



# -*- coding: utf-8 -*-
"""
obspy.realtime - Real time support for ObsPy
============================================

The obspy.realtime package extends the ObsPy core classes with real time
functionality.

:copyright:
    The ObsPy Development Team (devs@obspy.org), Anthony Lomax & Alessia Maggi
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)

The obspy.realtime development has been supported by the NERA project ["Network
of European Research Infrastructures for Earthquake Risk Assessment and
Mitigation" under the European Community's Seventh Framework Programme
(FP7/2007-2013) grant agreement n° 262330] and implemented within the
activities of the JRA2/WP12 "Tools for real-time seismology, acquisition and
mining".
"""
from obspy.realtime.rtmemory import RtMemory
from obspy.realtime.rttrace import RtTrace


if __name__ == '__main__':
    import doctest
    doctest.testmod(exclude_empty=True)



# -*- coding: utf-8 -*-

class TestCompatibility:
    def test_round(self):
        """
        Ensure round rounds correctly and returns expected data types.

        Maybe not needed after Py2 sunset? This usedd to be a test for
        compatibility function py3_round which was only an alias for round on
        Py3
        """
        # list of tuples; (input, ndigits, expected, excpected type)
        test_list = [
            (.222, 2, .22, float),
            (1516047903968282880, -3, 1516047903968283000, int),
            (1.499999999, None, 1, int),
            (.0222, None, 0, int),
            (12, -1, 10, int),
            (15, -1, 20, int),
            (15, -2, 0, int),
        ]
        for number, ndigits, expected, expected_type in test_list:
            out = round(number, ndigits)
            assert out == expected
            assert isinstance(out, expected_type)



# -*- coding: utf-8 -*-
"""
Module for handling station metadata

This module provides a class hierarchy to consistently handle station metadata.
This class hierarchy is closely modelled after the upcoming de-facto standard
format `FDSN StationXML <https://www.fdsn.org/xml/station/>`_ which was
developed as a human readable XML replacement for Dataless SEED.

.. note:: EarthScope/IRIS is maintaining a Java tool for converting dataless
          SEED into StationXML and vice versa at
          https://seiscode.iris.washington.edu/projects/stationxml-converter

:copyright:
    The ObsPy Development Team (devs@obspy.org) & Chad Trabant
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)

Reading
-------
StationXML files can be read using the
:func:`~obspy.core.inventory.inventory.read_inventory()` function that
returns an :class:`~obspy.core.inventory.inventory.Inventory` object.

>>> from obspy import read_inventory
>>> inv = read_inventory("/path/to/BW_RJOB.xml")
>>> inv  # doctest: +ELLIPSIS
<obspy.core.inventory.inventory.Inventory object at 0x...>
>>> print(inv)  # doctest: +NORMALIZE_WHITESPACE
Inventory created at 2013-12-07T18:00:42.878000Z
    Created by: fdsn-stationxml-converter/1.0.0
            http://www.iris.edu/fdsnstationconverter
    Sending institution: Erdbebendienst Bayern
    Contains:
        Networks (1):
            BW
        Stations (1):
            BW.RJOB (Jochberg, Bavaria, BW-Net)
        Channels (3):
            BW.RJOB..EHZ, BW.RJOB..EHN, BW.RJOB..EHE

The file format in principle is autodetected. However, the autodetection uses
the official StationXML XSD schema and unfortunately many real world files
currently show minor deviations from the official StationXML definition causing
the autodetection to fail. Thus, manually specifying the format is a good idea:

>>> inv = read_inventory("/path/to/BW_RJOB.xml", format="STATIONXML")

Class hierarchy
---------------
The :class:`~obspy.core.inventory.inventory.Inventory` class has a hierarchical
structure, starting with a list of :class:`Networks
<obspy.core.inventory.network.Network>`, each containing a list of
:class:`Stations <obspy.core.inventory.station.Station>` which again each
contain a list of :class:`Channels <obspy.core.inventory.channel.Channel>`.
The  :class:`Responses <obspy.core.inventory.response.Response>` are
attached to the channels as an attribute.

.. figure:: /_images/Inventory.png

>>> net = inv[0]
>>> net  # doctest: +ELLIPSIS
<obspy.core.inventory.network.Network object at 0x...>
>>> print(net)  # doctest: +NORMALIZE_WHITESPACE
Network BW (BayernNetz)
    Station Count: None/None (Selected/Total)
    -- - --
    Access: UNKNOWN
    Contains:
        Stations (1):
            BW.RJOB (Jochberg, Bavaria, BW-Net)
        Channels (3):
            BW.RJOB..EHZ, BW.RJOB..EHN, BW.RJOB..EHE


>>> sta = net[0]
>>> print(sta)  # doctest: +NORMALIZE_WHITESPACE
Station RJOB (Jochberg, Bavaria, BW-Net)
    Station Code: RJOB
    Channel Count: None/None (Selected/Total)
    2007-12-17T00:00:00.000000Z -
    Access: None
    Latitude: 47.7372, Longitude: 12.7957, Elevation: 860.0 m
    Available Channels:
        ..EH[ZNE]   200.0 Hz  2007-12-17(351) -

>>> cha = sta[0]
>>> print(cha)  # doctest: +NORMALIZE_WHITESPACE
Channel 'EHZ', Location ''
    Time range: 2007-12-17T00:00:00.000000Z - --
    Latitude: 47.7372, Longitude: 12.7957, Elevation: 860.0 m, Local Depth: 0.0 m
    Azimuth: 0.00 degrees from north, clockwise
    Dip: -90.00 degrees down from horizontal
    Channel types: TRIGGERED, GEOPHYSICAL
    Sampling Rate: 200.00 Hz
    Sensor (Description): Streckeisen STS-2/N seismometer (None)
    Response information available

>>> print(cha.response)  # doctest: +NORMALIZE_WHITESPACE +ELLIPSIS
Channel Response
   From M/S (Velocity in Meters Per Second) to COUNTS (Digital Counts)
   Overall Sensitivity: 2.5168e+09 defined at 0.020 Hz
   4 stages:
      Stage 1: PolesZerosResponseStage from M/S to V, gain: 1500
      Stage 2: CoefficientsTypeResponseStage from V to COUNTS, gain: 1.67...
      Stage 3: FIRResponseStage from COUNTS to COUNTS, gain: 1
      Stage 4: FIRResponseStage from COUNTS to COUNTS, gain: 1

Preview plots of station map and instrument response
----------------------------------------------------
For station metadata, preview plot routines for geographic location of stations
as well as bode plots for channel instrument response information are
available. The routines for station map plots are:

 * :meth:`Inventory.plot() <obspy.core.inventory.inventory.Inventory.plot>`
 * :meth:`Network.plot() <obspy.core.inventory.network.Network.plot>`

For example:

.. code-block:: python

    >>> from obspy import read_inventory
    >>> inv = read_inventory()
    >>> inv.plot()  # doctest: +SKIP

.. plot::

    from obspy import read_inventory
    inv = read_inventory()
    inv.plot()

The routines for bode plots of channel instrument response are:

 * :meth:`Inventory.plot_response()
   <obspy.core.inventory.inventory.Inventory.plot_response>`
 * :meth:`Network.plot_response()
   <obspy.core.inventory.network.Network.plot_response>`
 * :meth:`Station.plot() <obspy.core.inventory.station.Station.plot>`
 * :meth:`Channel.plot() <obspy.core.inventory.channel.Channel.plot>`
 * :meth:`Response.plot() <obspy.core.inventory.response.Response.plot>`

For example:

.. code-block:: python

    >>> from obspy import read_inventory
    >>> inv = read_inventory()
    >>> resp = inv[0][0][0].response
    >>> resp.plot(0.001, output="VEL")  # doctest: +SKIP

.. plot::

    from obspy import read_inventory
    inv = read_inventory()
    resp = inv[0][0][0].response
    resp.plot(0.001, output="VEL")

When plotting stations with instrumentation changes, epoch times can be added
to the legend labels:

.. code-block:: python

    >>> from obspy import read_inventory
    >>> inv = read_inventory()
    >>> inv = inv.select(station='RJOB', channel='EHZ')
    >>> inv.plot_response(0.001, label_epoch_dates=True)  # doctest: +SKIP

.. plot::

    from obspy import read_inventory
    inv = read_inventory()
    inv = inv.select(station='RJOB', channel='EHZ')
    inv.plot_response(0.001, label_epoch_dates=True)

For more examples see the :ref:`Obspy Gallery <gallery>`.

Dealing with the Response information
-------------------------------------
The :meth:`~obspy.core.inventory.response.Response.get_evalresp_response`
method will call some functions within evalresp to generate the response.

>>> response = cha.response
>>> response, freqs = response.get_evalresp_response(0.1, 16384, output="VEL")
>>> print(response)  # doctest: +NORMALIZE_WHITESPACE
 [  0.00000000e+00 +0.00000000e+00j  -1.36383361e+07 +1.42086194e+06j
   -5.36470300e+07 +1.13620679e+07j ...,   2.48907496e+09 -3.94151237e+08j
    2.48906963e+09 -3.94200472e+08j   2.48906430e+09 -3.94249707e+08j]


Some convenience methods to perform an instrument correction on
:class:`~obspy.core.stream.Stream` (and :class:`~obspy.core.trace.Trace`)
objects are available and most users will want to use those. The
:meth:`~obspy.core.stream.Stream.remove_response()` method deconvolves the
instrument response in-place. As always see the corresponding docs pages for a
full list of options and a more detailed explanation.

>>> from obspy import read
>>> st = read()
>>> inv = read_inventory("/path/to/BW_RJOB.xml")
>>> st.remove_response(
...     inventory=inv, output="VEL", water_level=20)  # doctest: +ELLIPSIS
<obspy.core.stream.Stream object at 0x...>

Writing
-------
:class:`~obspy.core.inventory.inventory.Inventory` objects can be exported to
StationXML files, e.g. after making modifications.

>>> inv.write('my_inventory.xml', format='STATIONXML')  # doctest: +SKIP
"""
# Don't change order! .util must be first.
from .util import (Angle, Azimuth, BaseNode, ClockDrift, Comment, Dip,
                   Distance, Equipment, ExternalReference, Frequency, Latitude,
                   Longitude, Operator, Person, PhoneNumber, SampleRate, Site)

from .channel import Channel
from .inventory import Inventory, read_inventory
from .network import Network
from .response import (CoefficientsTypeResponseStage,
                       CoefficientWithUncertainties, FilterCoefficient,
                       FIRResponseStage, InstrumentPolynomial,
                       InstrumentSensitivity, PolesZerosResponseStage,
                       PolynomialResponseStage, Response,
                       ResponseListResponseStage, ResponseStage)
from .station import Station


if __name__ == '__main__':
    import doctest
    doctest.testmod(exclude_empty=True)



# -*- coding: utf-8 -*-

import os


# generate credits
fh = open(os.path.join('source', 'credits.rst'), 'w', 'utf-8')
fh.write(""".. DON'T EDIT THIS FILE MANUALLY!
   Instead edit txt files in the credits folder and
   run ``make credits`` from command line to automatically create this file!

Contributors
============

We would like to thank our contributors, whose efforts make this software what
it is. These people have helped by writing code and documentation, and by
testing. They have created and maintained this product, its associated
libraries and applications, our build tools and our web sites.

.. rubric:: Hall of Fame

.. hlist::
    :columns: 3

""")


# add contributors
def name_key_function(name):
    """
    Key function for use with :py:func:`sorted` to sort full names given as
    "last name, first names" (e.g. `u'van Driel, Martin\n'`).
    """
    last, first = name.split(",")
    last = last.split()
    # "pop" any last name prefixes that should be ignored during sorting
    if last[0] in ["van"]:
        last.pop(0)
    return last, first


filename = os.path.join(os.pardir, os.pardir, 'CONTRIBUTORS.txt')
lines = [line for line in open(filename, 'r', 'utf-8').readlines()
         if line.strip()]
contributors = sorted(lines, key=name_key_function)

for item in contributors:
    fh.write("    * %s" % (item))

fh.write("""
.. rubric:: Funds

ObsPy was partially funded by the

""")

# add funds
filename = os.path.join('source', 'credits', 'FUNDS.txt')
funds = open(filename, 'r', 'utf-8').readlines()

for item in funds:
    fh.write("* %s" % (item))

fh.write("""
.. rubric:: Quotes

""")

# add quotes
filename = os.path.join('source', 'credits', 'QUOTES.txt')
funds = open(filename, 'r', 'utf-8').readlines()

for item in funds:
    item = item.split('---')
    fh.write("""
.. epigraph::
    %s""" % (item[0]))
    try:
        fh.write("""

    -- %s""" % (item[1]))
    except IndexError:
        pass

fh.close()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
obspy.io.ndk - NDK file support for ObsPy
=========================================

This module provides read support for the NDK files distributed by the
Global Centroid-Moment-Tensor (CMT) catalog.

:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)

The  `NDK
<https://www.ldeo.columbia.edu/~gcmt/projects/CMT/catalog/
allorder.ndk_explained>`_ format is used by the
`Global Centroid-Moment-Tensor (CMT) Project <http://www.globalcmt.org/>`_
and also in couple of other places to describe centroid moment tensor
solutions for earthquakes. Write support is not provided due to limited
potential use. Furthermore NDK files require some very specific information
about earthquakes which is not available all the timer


Acquiring NDK Files
-------------------

.. note::

    You always can either download the file and then pass the file location to
    ObsPy or you can directly pass the URL to ObsPy which will download it for
    you as demonstrated below.

The GCMT project offers two types of solutions: *Quick* (calculated within
hours of an event) and *Standard* (calculated later and with greater
accuracy). Quick solutions are available `individually
<https://www.ldeo.columbia.edu/~gcmt/projects/CMT/catalog/NEW_QUICK>`_ or
as a `single file
<https://www.ldeo.columbia.edu/~gcmt/projects/CMT/catalog/NEW_QUICK/qcmt.ndk>`_
collecting the most recent solutions. Standard solution are available via the
`CMT Catalog Files <http://www.globalcmt.org/CMTfiles.html>`_ website.


Basic Usage
-----------

.. note::

    During reading all values will be converted from ``dyne-cm`` to ``N m``.
    Furthermore the depth values will be converted to ``m`` and the source time
    function duration will be doubled to adhere to the QuakeML standard.  The
    region name in the NDK file will be replaced with its full Flinn-Engdahl
    region name using ObsPy's built-in region parser.

The :func:`~obspy.core.event.read_events` method is used to read NDK files to
:class:`~obspy.core.event.Catalog` objects.

>>> import obspy
>>> cat = obspy.read_events("/path/to/C200604092050A.ndk")
>>> print(cat)
1 Event(s) in Catalog:
2006-04-09T20:50:51.300000Z | -20.460,  -70.730 | 5.73 Mwc

One of the main purposes of this module is the conversion to QuakeML which
can easily be achieved with ObsPy.

>>> cat.write("C200604092050A.xml", format="quakeml")  # doctest: +SKIP

Instead of passing a filename it is also possible to specify a URL (in this
case to the solutions from February 2011).

>>> cat = obspy.read_events("https://www.ldeo.columbia.edu/~gcmt/projects/CMT/"
...                         "catalog/NEW_MONTHLY/2011/"
...                         "feb11.ndk")  # doctest: +SKIP
>>> print(cat)  # doctest: +SKIP
135 Event(s) in Catalog:
2011-02-01T03:32:08.600000Z | +22.520, +144.790 | 5.02 Mwc
2011-02-01T08:16:31.700000Z | +24.190, +121.730 | 5.33 Mwc
...
2011-02-28T20:45:45.700000Z | -20.520,  -69.440 | 5.27 Mwc
2011-02-28T23:42:21.500000Z | -29.650, -112.280 | 5.54 Mwc
To see all events call 'print(CatalogObject.__str__(print_all=True))'
>>> cat.plot()  # doctest: +SKIP

.. plot::

    import obspy
    cat = obspy.read_events("https://www.ldeo.columbia.edu/~gcmt/projects/CMT/"
                            "catalog/NEW_MONTHLY/2011/feb11.ndk")
    cat.plot()

It can also deal with large amounts of data, though it might take a while.

>>> cat = obspy.read_events("/path/to/jan76_dec10.ndk")  # doctest: +SKIP
>>> print(cat)  # doctest: +SKIP
33872 Event(s) in Catalog:
1976-01-01T01:29:53.400000Z | -29.250, -176.960 | 7.25  Mwc
1976-01-05T02:31:44.700000Z | -13.420,  -75.140 | 5.65  Mwc
...
2010-12-31T04:11:08.400000Z | -18.980, +167.590 | 5.07  Mwc
2010-12-31T16:30:59.700000Z |  +0.810,  -26.060 | 5.45  Mwc
To see all events call 'print(CatalogObject.__str__(print_all=True))'
>>> cat.plot()  # doctest: +SKIP

.. figure:: /_images/expensive_plots/GCMT_Catalog_1976-2010.png

"""
if __name__ == '__main__':
    import doctest
    doctest.testmod(exclude_empty=True)



# -*- coding: utf-8 -*-
"""
Directive for creating credits.rst page listing contributors, funds and quotes.
"""

import os


TEMPLATE_CONTRIBUTORS = """.. DON'T EDIT THIS FILE MANUALLY!
   Instead edit CONTRIBUTORS.txt, FUNDS.txt and QUOTES.txt.
   This file will be recreated automatically during building the docs.

Contributors
============

We would like to thank our contributors, whose efforts make this software what
it is. These people have helped by writing code and documentation, and by
testing. They have created and maintained this product, its associated
libraries and applications, our build tools and our web sites.

Hall of Fame
------------

.. hlist::
    :columns: 2

"""

TEMPLATE_FUNDS = """
Funds
-----

ObsPy was partially funded by the

"""

TEMPLATE_QUOTES = """
Quotes
------

"""


# add contributors
def _name_key_function(name):
    """
    Key function for use with :py:func:`sorted` to sort full names given as
    "last name, first names" (e.g. `u'van Driel, Martin\n'`).
    """
    last, first = name.split(",")
    last = last.split()
    # "pop" any last name prefixes that should be ignored during sorting
    if last[0] in ["van"]:
        last.pop(0)
    return last, first


def create_credits_page(app):
    # create credits.rst
    fh = open(os.path.join('source', 'credits.rst'),
              mode='w', encoding='utf-8', newline='\n')
    fh.write(TEMPLATE_CONTRIBUTORS)

    filename = os.path.join(os.pardir, os.pardir, 'obspy', 'CONTRIBUTORS.txt')
    lines = [line for line in open(filename, 'r', encoding='utf-8').readlines()
             if line.strip()]
    contributors = sorted(lines, key=_name_key_function)

    for item in contributors:
        fh.write("    * %s" % (item))
    fh.write(TEMPLATE_FUNDS)

    # add funds
    filename = os.path.join('source', 'credits', 'FUNDS.txt')
    funds = open(filename, 'r', encoding='utf-8').readlines()

    for item in funds:
        fh.write("* %s" % (item))
    fh.write(TEMPLATE_QUOTES)

    # add quotes
    filename = os.path.join('source', 'credits', 'QUOTES.txt')
    funds = open(filename, 'r', encoding='utf-8').readlines()

    for item in funds:
        item = item.split('---')
        # quote
        fh.write("\n\n    %s" % (item[0].strip()))
        # attribution, if given
        try:
            fh.write("\n\n    -- %s" % (item[1]))
        except IndexError:
            pass
    fh.close()


def setup(app):
    app.connect('builder-inited', create_credits_page)
    return {"parallel_write_safe": True,
            "parallel_read_safe": True}
